import numpy as np

from coolbox.utilities import get_logger
from coolbox.utilities.genome import GenomeRange

log = get_logger(__name__)


class PlotHist(object):
    """Mixin for plot Coverage plot(BigWig, BedGraph, BAM(coverage))."""

    def plot_hist(self, ax, gr: GenomeRange, indexes: np.ndarray, values: np.ndarray):
        assert len(indexes.shape) == 1 and 1 <= len(values.shape) <= 2
        style = self.properties.get('style', 'line')
        if 'heatmap' in style:
            self.plot_heatmap(ax, gr, values)
        elif 'fill' in style:
            self.plot_fill(ax, indexes, values)
        elif 'scatter' in style:
            self.plot_scatter(ax, indexes, values)
        elif 'line' in style:
            self.plot_line(ax, indexes, values)

        ymin, ymax = self.adjust_plot(ax, gr)
        # disable plot data range in coverage mode
        if self.properties['data_range_style'] != 'no' and 'track' not in self.__dict__:
            self.plot_data_range(ax, ymin, ymax, self.properties['data_range_style'], gr)

    def plot_heatmap(self, ax, gr: GenomeRange, mat: np.ndarray):
        # no threshold supported
        cmap = self.properties.get('cmap', None)
        if len(mat.shape) == 1:
            height = self.properties.get('height', 3)
            mat = np.array([mat] * height)
        ax.matshow(mat, cmap=cmap, aspect="auto", extent=(gr.start, gr.end, 0, mat.shape[0]))

    def plot_fill(self, ax, indexes, values):
        alpha = self.properties.get('alpha', 1.0)
        threshold = self.properties.get("threshold", 0)
        ax.fill_between(indexes, 0, values, where=(values > threshold),
                        linewidth=0.1, color=self.properties['threshold_color'],
                        facecolor=self.properties['threshold_color'],
                        alpha=alpha)
        ax.fill_between(indexes, 0, values, where=(values < threshold),
                        linewidth=0.1, color=self.properties['color'],
                        facecolor=self.properties['color'],
                        alpha=alpha)

    def plot_scatter(self, ax, indexes: np.ndarray, values: np.ndarray):
        threshold = self.properties.get("threshold", float("inf"))
        color = self.properties['color']
        size = self.properties['size']
        alpha = self.properties['alpha']
        mask = values > threshold
        ax.scatter(indexes[mask], values[mask], s=size, alpha=alpha, c=self.properties.get('threshold_color', color))
        mask = ~mask
        ax.scatter(indexes[mask], values[mask], s=size, alpha=alpha, c=color)

    def plot_line(self, ax, indexes, values):
        # reference for plotting with threshold: https://stackoverflow.com/a/30122991/10336496
        # if the data is 2d matrix, transpose to comply wtih matplotlib
        if len(values.shape) == 2:
            values = values.T
        fmt = self.properties.get("fmt")
        line_width = self.properties.get('line_width', 1)
        color = self.properties.get('color')
        alpha = self.properties.get("alpha", 1.0)
        threshold = float(self.properties.get("threshold"))
        threshold_color = self.properties.get("threshold_color")
        ax.plot(indexes, values, fmt, linewidth=line_width, color=color, alpha=alpha)
        if threshold and np.sum(values > threshold) > 0:
            masked_values = np.ma.masked_greater_equal(values, threshold)
            ax.plot(indexes, masked_values, fmt, linewidth=line_width, color=threshold_color, alpha=alpha)

    def plot_data_range(self, ax, ymin, ymax, data_range_style, gr: GenomeRange):
        if data_range_style == 'text':
            self.plot_text_range(ax, ymin, ymax, gr)
        else:  # 'y-axis' style
            try:
                y_ax = self.y_ax
                self.plot_yaxis_range(ax, y_ax)
            except AttributeError as e:
                log.exception(e)
                msg = "If use y-axis style data range must, must set the .y_ax attribute, switch to text style."
                log.warning(msg)
                self.plot_data_range(ax, ymin, ymax, 'text', gr)

    def plot_yaxis_range(self, plot_axis, y_ax):
        """
        Plot the scale of the y axis with respect to the plot_axis

        plot something that looks like this:
        ymax ┐
             │
             │
        ymin ┘

        Parameters
        ----------
        plot_axis : matplotlib.axes.Axes
            Main plot axis.

        y_ax : matplotlib.axes.Axes
            Axis to use to plot the scale
        """

        if 'show_data_range' in self.properties and self.properties['show_data_range'] == 'no':
            return

        def value_to_str(value):
            if value % 1 == 0:
                str_value = str(int(value))
            else:
                if value < 0.01:
                    str_value = "{:.4f}".format(value)
                else:
                    str_value = "{:.2f}".format(value)
            return str_value

        ymin, ymax = plot_axis.get_ylim()

        ymax_str = value_to_str(ymax)
        ymin_str = value_to_str(ymin)
        x_pos = [0, 0.5, 0.5, 0]
        y_pos = [0.01, 0.01, 0.99, 0.99]
        y_ax.plot(x_pos, y_pos, color='black', linewidth=1, transform=y_ax.transAxes)
        y_ax.text(-0.2, -0.01, ymin_str, verticalalignment='bottom', horizontalalignment='right',
                  transform=y_ax.transAxes)
        y_ax.text(-0.2, 1, ymax_str, verticalalignment='top', horizontalalignment='right', transform=y_ax.transAxes)
        y_ax.patch.set_visible(False)

    def plot_text_range(self, ax, ymin, ymax, gr: GenomeRange):
        ydelta = ymax - ymin

        # set min max
        format_lim = lambda lim: int(lim) if float(lim) % 1 == 0 else "{:.2f}".format(lim)
        ymax_print = format_lim(ymax)
        ymin_print = format_lim(ymin)
        small_x = 0.01 * gr.length
        # by default show the data range
        ax.text(gr.start - small_x, ymax - ydelta * 0.2,
                "[ {} ~ {} ]".format(ymin_print, ymax_print),
                horizontalalignment='left',
                verticalalignment='top')

    def adjust_plot(self, ax, gr: GenomeRange):
        ax.set_xlim(gr.start, gr.end)
        ymin, ymax = ax.get_ylim()
        if 'max_value' in self.properties and self.properties['max_value'] != 'auto':
            ymax = self.properties['max_value']
        if 'min_value' in self.properties and self.properties['min_value'] != 'auto':
            ymin = self.properties['min_value']

        if 'orientation' in self.properties and self.properties['orientation'] == 'inverted':
            ax.set_ylim(ymax, ymin)
        else:
            ax.set_ylim(ymin, ymax)
        return ymin, ymax



import pandas as pd
from matplotlib.collections import BrokenBarHCollection

from coolbox.utilities import (
    get_logger, GenomeRange, file_to_intervaltree, hex2rgb,
)
from .base import Track

log = get_logger(__name__)


# TODO change properties to new mode
class Ideogram(Track):
    """
    The chromosome ideograme track.

    Parameters
    ----------
    file_ : str
        Path to chromosome ideogram txt file,
        ideogram file is download from the UCSC Table Browser CytoBandIdeo table (in "all table" group).
        see: http://genome.ucsc.edu/cgi-bin/hgTables?hgta_group=allTables&hgta_table=cytoBandIdeo

    color_scheme : dict, optional
        Color scheme of ideogram, default: Ideogram.DEFAULT_COLOR_SCHEME

    show_name : bool, optional
        Show band name or not. default True.

    font_size : int, optional
        Band name font size.

    border_color : str, optional
        Border color. default: '#000000'

    border_width : float, optional
        Border width. default: 1.2
    """

    DEFAULT_HEIGHT = 1.2
    DEFAULT_COLOR_SCHEME = {
        'gneg': '#ffffff',
        'gpos25': '#999999',
        'gpos50': '#666666',
        'gpos75': '#333333',
        'gpos100': '#000000',
        'acen': '#cc6666',
        'gvar': '#cccccc',
        'stalk': '#e5e5e5',
    }
    DEFAULT_FONT_SIZE = 12
    DEFAULT_BORDER_WIDTH = 1.2

    def __init__(self, file_, **kwargs):
        properties_dict = {
            'file': file_,
            'color_scheme': Ideogram.DEFAULT_COLOR_SCHEME,
            'show_name': True,
            'font_size': Ideogram.DEFAULT_FONT_SIZE,
            'border_color': '#000000',
            'border_width': Ideogram.DEFAULT_BORDER_WIDTH,
            'height': Ideogram.DEFAULT_HEIGHT,
            'title': '',
        }
        properties_dict.update(kwargs)
        super().__init__(properties_dict)
        self.file = self.properties['file']
        self.interval_tree, _, _ = file_to_intervaltree(self.file)

    def lookup_band_color(self, band_type):
        color_scheme = self.properties['color_scheme']
        if band_type in color_scheme:
            return color_scheme[band_type]
        else:
            return color_scheme['gneg']

    def fetch_data(self, gr: GenomeRange, **kwargs):
        if gr.chrom not in self.interval_tree:
            gr.change_chrom_names()
        bands_in_region = sorted(self.interval_tree[gr.chrom][gr.start:gr.end])
        rows = []
        for itv in bands_in_region:
            start, end = itv.begin, itv.end
            band_name, band_type = itv.data[:2]
            rows.append([gr.chrom, start, end, band_name, band_type])
        fields = ['chrom', 'start', 'end', 'name', 'gieStain']
        df = pd.DataFrame(rows, columns=fields)
        return df

    def plot(self, ax, gr: GenomeRange, **kwargs):
        self.ax = ax
        df = self.fetch_data(gr)
        xranges, colors = [], []
        band_height = self.properties['height']
        for _, row in df.iterrows():
            start, end = row['start'], row['end']
            band_name, band_type = row['band_name'], row['band_type']
            band_color = self.lookup_band_color(band_type)
            xranges.append((start, end))
            colors.append(band_color)
            if self.properties['show_band_name'] != 'no':
                if gr.length < 80_000_000:
                    self.plot_text(band_name, start, end, band_color)
        coll = BrokenBarHCollection(xranges, (0, band_height), facecolors=colors,
                                    linewidths=self.properties['border_width'],
                                    edgecolors=self.properties['border_color'])
        ax.add_collection(coll)
        ax.set_ylim(-0.1, band_height + 0.1)
        ax.set_xlim(gr.start, gr.end)
        self.plot_label()

    def plot_text(self, band_name, start, end, band_color):
        band_height = self.properties['height']
        x_pos = start + (end - start) * 0.15
        y_pos = band_height / 2
        if isinstance(band_color, str):
            rgb = hex2rgb(band_color)
        else:
            rgb = band_color
        if sum(rgb) < 100:
            color = '#e2e2e2'
        else:
            color = '#000000'
        self.ax.text(x_pos, y_pos, band_name, fontsize=self.properties['font_size'], color=color)



from .base import ArcsBase
from .bedpe import BEDPE
from .hicpeaks import HiCPeaks
from .pairs import Pairs


def Arcs(file_, *args, **kwargs) -> ArcsBase:
    """Compose BEDPE, Pairs track automatically based on type of file extension.(.bedpe, .pairs)
    """
    if file_.endswith((".bedpe", ".bedpe.bgz")):
        return BEDPE(file_, *args, **kwargs)
    elif file_.endswith((".pairs", ".pairs.bgz")):
        return Pairs(file_, *args, **kwargs)
    else:
        raise NotImplementedError("Arcs track only support .bedpe or .pairs input format.")



import os.path as osp
import subprocess as subp

import numpy as np
import pandas as pd

from coolbox.utilities import GenomeRange
from coolbox.utilities.bed import tabix_query, build_snp_index
from .base import HistBase


class SNP(HistBase):
    """
    Track for show SNPs Manhattan plot.

    Input file is a tab-split file contains SNP's chrom, position, pvalue information.
    You should specify it's column indexes by `col_chrom`, `col_pos`, `col_pval` arguments.

    Parameters
    ----------
    file : str
        Path to input .snp/.vcf file.

    col_chrom : int
        Column index of seqname(chromosome).

    col_pos : int
        Column index of snp position.

    col_pval : int
        Column index of snp p-value.

    pval_transform : {'-log2', '-log10'}
        Transform the p value. Default '-log10'.
    """
    COL_CHROM = 0
    COL_POS = 2
    COL_PVAL = 9

    DEFAULT_PROPERTIES = {
        "style": HistBase.STYLE_SCATTER,
        "color": "grey",
        "threshold_color": "#ff9c9c",
        "threshold": 0.05,
        "alpha": 0.5,
        "size": 10,
        # TODO if min_value is set to 'auto', the min_value is not 0, different to the original codes
        "min_value": "0",
        "max_value": "auto",
        "height": 5.0,
        # processing
        "pval_transform": "-log10",
        "col_chrom": COL_CHROM,
        "col_pos": COL_POS,
        "col_pval": COL_PVAL,
    }

    def __init__(self, file, **kwargs):
        properties = SNP.DEFAULT_PROPERTIES.copy()
        properties.update({
            'file': file,
            **kwargs
        })
        super().__init__(**properties)
        self.bgz_file = build_snp_index(
            file,
            self.properties['col_chrom'],
            self.properties['col_pos']
        )
        # TODO what does this mean?
        self.properties['threshold'] = self.transform_fn()(self.properties['threshold'])

    def fetch_plot_data(self, gr: GenomeRange, **kwargs):
        df = self.fetch_data(gr, **kwargs)
        df['score'] = self.transform_fn()(df['score'])
        return df

    def fetch_data(self, gr: GenomeRange, **kwargs):
        ix_chrom = self.properties['col_chrom']
        ix_pos = self.properties['col_pos']
        ix_pval = self.properties['col_pval']
        rows = self.load_range(gr)
        if len(rows) == 0:
            gr.change_chrom_names()
            rows = self.load_range(gr)
        df = pd.DataFrame(rows)
        if df.shape[0] > 0:
            columns = [f'col_{i}' for i in range(df.shape[1])]
            columns[ix_chrom] = "chrom"
            columns[ix_pos] = "pos"
            columns[ix_pval] = "score"
            df.columns = columns
        return df

    def load_range(self, gr):
        rows = []
        ix_pos = self.properties['col_pos']
        ix_pval = self.properties['col_pval']
        for items in tabix_query(self.bgz_file, gr.chrom, gr.start, gr.end):
            items[ix_pos] = int(items[ix_pos])
            items[ix_pval] = float(items[ix_pval])
            rows.append(items)
        return rows

    def transform_fn(self):
        method = self.properties.get('pval_transform', '-log2')
        if method == "-log2":
            return lambda x: -np.log2(x)
        elif method == "-log10":
            return lambda x: -np.log10(x)
        else:
            return lambda x: x



from coolbox.utilities import (
    opener, ReadBed,
    Interval, IntervalTree,
    rgb2hex,
    get_logger, GenomeRange,
    to_gr
)
from .base import Coverage

log = get_logger(__name__)


class _Highlights(object):
    DEFAULT_COLOR = '#ff5d0f'

    def fetch_data(self, gr: GenomeRange, **kwargs):
        gr = to_gr(gr)
        regions = []

        if gr.chrom not in list(self.interval_tree):
            gr.change_chrom_names()

        for region in sorted(self.interval_tree[gr.chrom][gr.start - 10000:gr.end + 10000]):
            regions.append((region.begin, region.end, region.data))

        return regions

    def plot(self, ax, gr: GenomeRange, **kwargs):

        regions = self.fetch_data(gr, **kwargs)

        for (start, end, color) in regions:
            if self.properties['color'] != 'bed_rgb':
                color = self.properties['color']
            if type(color) is not str:
                color = rgb2hex(*color)

            ax.axvspan(start, end, color=color, alpha=self.properties['alpha'])

            if self.properties['border_line'] == 'yes':
                # plot border line
                ymin, ymax = ax.get_ylim()
                ax.vlines([start, end], ymin, ymax,
                          linestyle=self.properties['border_line_style'],
                          linewidth=self.properties['border_line_width'],
                          color=self.properties['border_line_color'],
                          alpha=self.properties['border_line_alpha'])


class HighLightsFromFile(Coverage, _Highlights):
    """
    High light regions coverage, read the regions from the file.

    Parameters
    ----------
    file_ : str
        Path to the file.

    color : str, optional
        High light region color,
        use 'bed_rgb' for specify color from the file, default 'bed_rgb'.

    alpha : float, optional
        High light region alpha value, default 0.1.

    border_line : bool, optional
        Plot border line or not, default True.

    border_line_style : str, optional
        Border line style, default 'dashed'.

    border_line_width : float, optional
        Border line width, default 1.0.

    border_line_color : str, optional
        Border line color, default '#000000'

    border_line_alpha : float, optional
        Border line alpha value, default 0.8

    name : str, optional
        The name of thr Coverage.
    """

    def __init__(self, file_, **kwargs):
        properties_dict = {
            "file": file_,
            "color": "bed_rgb",
            "alpha": 0.1,
            "border_line": True,
            "border_line_style": "dashed",
            "border_line_width": 0,
            "border_line_color": "#000000",
            "border_line_alpha": 0.8,
        }
        properties_dict.update(kwargs)
        super().__init__(properties_dict)
        self.interval_tree = self.__process_bed()

    def __process_bed(self):
        bed = ReadBed(opener(self.properties['file']))

        if self.properties['color'] == 'bed_rgb' and bed.file_type not in ['bed12', 'bed9']:
            log.warning("*WARNING* Color set to 'bed_rgb', but bed file does not have the rgb field. The color has "
                        "been set to {}".format(HighLights.DEFAULT_COLOR))
            self.properties['color'] = HighLights.DEFAULT_COLOR

        interval_tree = {}

        for intval in bed:

            if intval.chromosome not in interval_tree:
                interval_tree[intval.chromosome] = IntervalTree()

            if self.properties['color'] == 'bed_rgb':
                color = intval.rgb
            else:
                color = self.properties['color']
            interval_tree[intval.chromosome].add(Interval(intval.start, intval.end, color))

        return interval_tree


class HighLights(Coverage, _Highlights):
    """
    High light region.

    Parameters
    ----------
    highlight_regions : list of {str, tuple}
        A list of regions for highlights, region can be expressed as a tuple or string.
        region tuple like:
        [('chr1', 100000, 120000), ('chr2', 130000, 150000)]
        region string format: `chr:start-end` like:
        ['chr1:100000-120000', 'chr2:130000-150000'].

    color : str, optional
        High light region color, default HighLights.DEFAULT_COLOR.

    alpha : float, optional
        High light region alpha value, default 0.5

    border_line : bool, optional
        Plot border line or not, default True.

    border_line_style : str, optional
        Border line style, default 'dashed'

    border_line_width : float, optional
        Border line width, default 1.0

    border_line_color : str, optional
        Border line color, default '#000000'

    border_line_alpha : float, optional
        Border line alpha value, default 0.8

    name : str, optional
        The name of thr Coverage.
    """

    DEFAULT_COLOR = "#ff9c9c"

    def __init__(self, highlight_regions, **kwargs):
        if not isinstance(highlight_regions, list):
            highlight_regions = [highlight_regions]
        properties_dict = {
            "highlight_regions": highlight_regions,
            "color": HighLights.DEFAULT_COLOR,
            "alpha": 0.25,
            "border_line": True,
            "border_line_style": "dashed",
            "border_line_width": 0,
            "border_line_color": "#000000",
            "border_line_alpha": 0.8,
        }
        properties_dict.update(kwargs)
        super().__init__(properties_dict)
        self.interval_tree = self.__intervaltree_from_list(self.properties['highlight_regions'])

    # TODO  may be duplicate of vlines's method
    def __intervaltree_from_list(self, region_list):
        from intervaltree import IntervalTree
        itree = {}
        for r in region_list:
            if isinstance(r, str):
                grange = GenomeRange(r)
            elif isinstance(r, tuple):
                grange = GenomeRange(r[0], r[1], r[2])
            elif isinstance(r, GenomeRange):
                grange = r
            else:
                raise ValueError("position must be a tuple or string.")
            chr_ = grange.chrom
            itree.setdefault(chr_, IntervalTree())
            itree[chr_][grange.start:grange.end + 1] = grange
        return itree



from typing import Union, Tuple

import numpy as np
import pandas as pd

from coolbox.utilities import (
    split_genome_range, change_chrom_names,
    GenomeRange, get_logger,
)
from coolbox.utilities.bed import tabix_query, build_bedgraph_bgz
from .base import HistBase

log = get_logger(__name__)


class BedGraph(HistBase):
    """
    BedGraph track.

    Parameters
    -----------
    file : str
        File path of bedgraph file.

    """

    DEFAULT_PROPERTIES = {
        "style": HistBase.STYLE_FILL,
    }

    def __init__(self, file, **kwargs):
        properties = BedGraph.DEFAULT_PROPERTIES.copy()
        properties.update({
            'file': file,
            **kwargs
        })
        super().__init__(**properties)
        self.bgz_file = build_bedgraph_bgz(file)

    def fetch_plot_data(self, gr: GenomeRange, **kwargs) -> pd.DataFrame:
        itv_df = self.fetch_data(gr, **kwargs)
        index_array = itv_df['start'] + (itv_df['end'] - itv_df['start']) / 2
        itv_df['pos'] = index_array
        return itv_df

    def fetch_data(self, gr: GenomeRange, **kwargs) -> pd.DataFrame:
        rows = self.load(gr)
        if len(rows) == 0:
            gr.chrom = change_chrom_names(gr.chrom)
            rows = self.load(gr)

        intval_table = pd.DataFrame(rows, columns=['chromsome', 'start', 'end', 'score'])

        return intval_table

    def load(self, genome_range):
        rows = []
        gr = genome_range
        for it in tabix_query(self.bgz_file, gr.chrom, gr.start, gr.end, split=True):
            rows.append([
                it[0], int(it[1]), int(it[2]), float(it[3])
            ])
        return rows



import abc


class WidgetsBox(abc.ABC):
    """
    Widgets panel base class.
    """

    def __init__(self, *args, **kwargs):
        """
        Parameters
        ----------
        browser : BrowserBase
            Browser object.

        frame_widget : ipywidgets.Image
            frame widget object.
        """
        self.browser = args[0]
        self.frame_widget = args[1]

        self.widgets_dict = self.get_widgets_dict()
        self.refresh_widgets()
        self.panel = self.compose_panel(self.widgets_dict)
        self.register_events_handler()

    def get_widgets_dict(self):
        pass

    def refresh_widgets(self):
        pass

    def compose_panel(self, widgets_dict):
        pass

    def register_events_handler(self):
        pass



import numpy as np

from coolbox.utilities.genome import GenomeRange
from coolbox.core.track.hicmat.base import HicMatBase
from coolbox.core.track.hicmat.hicmat import HiCMat


class HiCDiff(HicMatBase):
    """
    Track for express the comparison between two HiC Track.

    Parameters
    ----------
    hic1 : coolbox.api.track.Cool
        First HiC Track or hic file path(.cool, .mcool, .hic).

    hic2 : coolbox.api.track.Cool
        Second HiC Track or hic file path(.cool, .mcool, .hic).

    args_hic: dict
        Key word arguments send to create Cool/DoctHic instance if the input hic1/hic2 is file.

    diff_method : str
        Difference method ('diff', 'log2fc'), default 'diff'


    """

    DEFAULT_ARGS_HIC = {
        "transform": "no",
        "normalize": "expect",
        "gaussian_sigma": "no",
        "process_func": "no",
    }

    DEFAULT_PROPERTIES = {
        'style': HicMatBase.STYLE_WINDOW,
        "cmap": "bwr",
        "transform": False,
        'diff_method': "diff",
        'args_hic': DEFAULT_ARGS_HIC,
    }

    def __init__(self, hic1, hic2, **kwargs):
        properties = HiCDiff.DEFAULT_PROPERTIES.copy()
        properties.update(kwargs)
        super().__init__(**properties)

        # TODO how to set hic_args in cli mode ?
        if isinstance(hic1, str):
            hic1 = HiCMat(hic1)
        if isinstance(hic2, str):
            hic2 = HiCMat(hic2)
        self.properties.update({
            'hic1': hic1,
            'hic2': hic2,
        })

        self.mat1 = None
        self.mat2 = None

    def fetch_data(self, gr: GenomeRange, **kwargs) -> np.ndarray:
        # TODO find a better way
        # For cases when updating of Frame's depth_ratio leads to un equal matrix sizes for different tracks).
        args_hic = self.properties.copy()
        args_hic.update(self.properties.get('args_hic', {}))
        for hic in ('hic1', 'hic2'):
            self.properties[hic].properties.update(args_hic)

        hic1: HicMatBase = self.properties['hic1']
        hic2: HicMatBase = self.properties['hic2']
        # must set to avoid re-change GenomeRange in window style
        kwargs['gr_updated'] = True
        # use transformed gr for cases in 'window' style, other wise the gr would be transformed two times
        self.mat1 = mat1 = hic1.fetch_plot_data(gr, **kwargs)
        self.mat2 = mat2 = hic2.fetch_plot_data(gr, **kwargs)
        diff_mat = self.diff_matrix(mat1, mat2)
        try:
            self.SMALL_VALUE = diff_mat[diff_mat > 0].min()
        except Exception:
            pass
        return diff_mat

    def diff_matrix(self, mat1, mat2):
        diff_mth = self.properties['diff_method']
        if diff_mth == 'log2fc':
            return np.log2((mat1 + 1) / (mat2 + 1))
        else:
            return mat1 - mat2



import abc
from typing import Union
from collections import OrderedDict
from copy import copy

from coolbox.utilities import (
    GenomeRange,
    op_err_msg,
    get_logger,
)

log = get_logger(__name__)

GR = Union[str, GenomeRange]


class FrameBase(abc.ABC):

    def __init__(self, properties_dict, *args, **kwargs):
        # init range
        if 'gr' in kwargs:
            range_ = kwargs['gr']
            if isinstance(range_, GenomeRange):
                self.current_range = range_
            else:
                # init from genome range string
                # e.g. `frame = Frame(gr="chr1:1000-2000")`
                self.current_range = GenomeRange(range_)
        else:
            self.current_range = None
        if 'gr2' in kwargs:
            range_ = kwargs['gr2']
            if isinstance(range_, GenomeRange):
                self.current_range = range_
            else:
                # init from genome range string
                # e.g. `frame = Frame(gr="chr1:1000-2000")`
                self.current_range = GenomeRange(range_)
        else:
            self.current_range2 = None

        self.properties = properties_dict

    @abc.abstractmethod
    def plot(self):
        pass

    @abc.abstractmethod
    def show(self):
        pass

    @property
    @abc.abstractmethod
    def tracks(self):
        pass

    @abc.abstractmethod
    def add_track(self, track):
        pass

    def goto(self, gr: GR = None, gr2: GR = None):
        """
        Go to the range on the genome.

        Parameters
        ----------
        gr2
        gr : {str, GenomeRange}
            The range string,
            like "chr1:1000000-2000000", or GenomeRange object.

        Examples
        --------
        >>> frame = Frame()
        >>> frame.goto("chrX:3000000-5000000")
        >>> str(frame.current_range)
        'chrX:3000000-5000000'
        >>> frame.goto(GenomeRange("chr1", 1000, 2000))
        >>> str(frame.current_range)
        'chr1:1000-2000'
        """
        if gr:
            self.current_range = gr if isinstance(gr, GenomeRange) else GenomeRange(gr)
        if gr2:
            self.current_range2 = gr if isinstance(gr2, GenomeRange) else GenomeRange(gr2)

    def fetch_data(self, gr: GR = None, gr2: GR = None):
        self.goto(gr, gr2)
        gr, gr2 = self.current_range, self.current_range2
        if gr is None:
            raise ValueError("No GenomeRange history found.")

        tracks_data = OrderedDict()
        for name, track in self.tracks.items():
            if hasattr(track, 'fetch_data'):
                data = track.fetch_data(gr, gr2=gr2)
            else:
                data = []
            tracks_data.update([(name, data)])

        return tracks_data

    def fetch_plot_data(self, gr: GR = None, gr2: GR = None):
        self.goto(gr, gr2)
        gr, gr2 = self.current_range, self.current_range2
        if gr is None:
            raise ValueError("No GenomeRange history found.")

        tracks_data = OrderedDict()
        for name, track in self.tracks.items():
            if hasattr(track, 'fetch_plot_data'):
                data = track.fetch_plot_data(gr, gr2=gr2)
            else:
                data = []
            tracks_data.update([(name, data)])

        return tracks_data

    def add_feature_to_tracks(self, feature):
        """
        Add feature to all tracks in this frame.

        Parameters
        ----------
        feature : Feature
            Feature object to be added to Frame's tracks.

        Examples
        --------
        >>> from coolbox.core.feature import Color
        >>> frame = Frame()
        >>> frame.add_feature_to_tracks(Color('#66ccff'))
        >>> assert all([track.properties['color'] == '#66ccff' for track in frame.tracks.values()])
        """
        from ..feature import Feature
        assert isinstance(feature, Feature), "feature must a Feature object."
        for track in self.tracks.values():
            track.properties.update(feature.properties)

    def add_cov_to_tracks(self, cov):
        """
        Add coverage to all tracks in this frame.

        Parameters
        ----------
        cov : Coverage
            Coverage object to be added to Frame's tracks.

        Examples
        --------
        >>> from coolbox.core.track import XAxis, BigWig
        >>> from coolbox.core.coverage import HighLights
        >>> frame = Frame()
        >>> frame.add_track(XAxis())
        >>> frame.add_track(BigWig("tests/test_data/bigwig_chrx_2e6_5e6.bw"))
        >>> highlights = HighLights([("chr1", 1000, 2000), ("chr2", 3000, 4000)])
        >>> frame.add_cov_to_tracks(highlights)
        >>> assert [track.coverages[0] is highlights for track in frame.tracks.values()]
        """
        for track in self.tracks.values():
            track.append_coverage(cov)

    def set_tracks_min_max(self, min_, max_, name=None):
        """
        set all bigwig and bedgraph tracks's min and max value.

        Examples
        --------
        >>> frame = Frame()
        >>> frame.set_tracks_min_max(-10, 10)
        >>> assert all([track.properties['min_value'] == -10 for track in frame.tracks.values()])
        >>> assert all([track.properties['min_value'] ==  10 for track in frame.tracks.values()])
        """
        from ..track import BedGraph, BigWig
        if name is None:  # set all setable tracks
            for track in self.tracks.values():
                if isinstance(track, BedGraph) or \
                        isinstance(track, BigWig):
                    track.properties['min_value'] = min_
                    track.properties['max_value'] = max_
        else:  # set specified track
            if name not in self.tracks:
                log.warning("Track {name} not in frame")
                return
            track = self.tracks[name]
            if 'min_value' in track.properties:
                track.properties['min_value'] = min_
                track.properties['max_value'] = max_

    def __add__(self, other):
        """
        Examples
        --------
        >>> from coolbox.api import *
        >>> frame1 = Frame()
        >>> frame2 = Frame()

        Rule: Frame + Track == Frame
        >>> frame3 = (frame1 + XAxis())
        >>> isinstance(frame3, Frame)
        True
        >>> len(frame3.tracks)
        1

        Rule: Frame + Frame == Frame
        >>> isinstance(frame1 + frame2, Frame)
        True

        Rule: Frame + Feature == Frame
        >>> f = XAxis() + BigWig("tests/test_data/bigwig_chrx_2e6_5e6.bw")
        >>> f = f + Color('#66ccff')
        >>> isinstance(f, Frame)
        True
        >>> 'color' not in list(f.tracks.values())[0].properties
        True
        >>> list(f.tracks.values())[1].properties['color']
        '#66ccff'

        Rule: Frame + Coverage == Frame
        >>> f = XAxis() + BigWig("tests/test_data/bigwig_chrx_2e6_5e6.bw")
        >>> highlight = HighLights([("chr1", 1000, 2000), ("chr2", 3000, 4000)])
        >>> f = f + highlight
        >>> isinstance(f, Frame)
        True
        >>> len(list(f.tracks.values())[0].coverages)
        0
        >>> list(f.tracks.values())[1].coverages[0] is highlight
        True

        Rule: Frame + CoverageStack == Frame
        >>> cov_stack = HighLights([("chr1", 1000, 2000)]) + HighLights([("chr1", 4000, 6000)])
        >>> f = XAxis() + XAxis()
        >>> f = f + cov_stack
        >>> isinstance(f, Frame)
        True
        >>> len(list(f.tracks.values())[0].coverages)
        0
        >>> len(list(f.tracks.values())[1].coverages)
        2

        Rule: Frame + WidgetsPanel == Browser
        >>> f = XAxis() + XAxis()
        >>> bsr = f + WidgetsPanel()
        >>> isinstance(bsr, Browser)
        True

        >>> f + 1 # error operation
        Traceback (most recent call last):
        ...
        TypeError: unsupported operand type(s) for +: '<class 'coolbox.core.frame.Frame'>' and '<class 'int'>'


        """
        from ..track.base import Track
        from ..feature import Feature, FrameFeature
        from ..coverage.base import Coverage, CoverageStack
        from ..browser import WidgetsPanel, Browser

        result = copy(self)

        if isinstance(other, Track):
            result.add_track(other)
            return result
        elif isinstance(other, FrameBase):
            for track in other.tracks.values():
                result.add_track(track)
            result.properties.update(other.properties)
            return result
        elif isinstance(other, FrameFeature):
            result.properties.update(other.properties)
            return result
        elif isinstance(other, Feature):
            if len(result.tracks) != 0:
                last = list(result.tracks.values())[-1]
                last.properties.update(other.properties)
            return result
        elif isinstance(other, Coverage):
            if len(result.tracks) != 0:
                last = list(result.tracks.values())[-1]
                last.append_coverage(other)
            return result
        elif isinstance(other, CoverageStack):
            if len(result.tracks) != 0:
                last = list(result.tracks.values())[-1]
                last.pile_coverages(other.coverages, pos='top')
            return result
        elif isinstance(other, WidgetsPanel):
            return Browser(self, reference_genome=other.ref, widgets_box=other.type)
        else:
            raise TypeError(op_err_msg(self, other))

    def __mul__(self, other):
        """
        Examples
        --------
        >>> from coolbox.core.track import XAxis, BigWig
        >>> from coolbox.core.coverage import HighLights
        >>> from coolbox.core.feature import Color

        Rule: Frame * Feature == Frame
        >>> f = XAxis() + BigWig("tests/test_data/bigwig_chrx_2e6_5e6.bw")
        >>> f = f * Color("#ff9c9c")
        >>> assert all([track.properties['color'] == '#ff9c9c' for track in f.tracks.values()])

        Rule: Frame * Coverage == Frame
        >>> cov = HighLights([('chr1', 1000, 2000)])
        >>> f = f * cov
        >>> assert all([track.coverages[0] is cov for track in f.tracks.values()])

        """
        from ..coverage.base import Coverage
        from ..feature import Feature

        if isinstance(other, Coverage):
            result = copy(self)
            result.add_cov_to_tracks(other)
            return result
        elif isinstance(other, Feature):
            result = copy(self)
            result.add_feature_to_tracks(other)
            return result
        else:
            raise TypeError(op_err_msg(self, other, op='*'))


if __name__ == "__main__":
    import doctest

    doctest.testmod()



import copy
import math

import matplotlib.colors as colors
import matplotlib.pyplot as plt
import numpy as np
from matplotlib import transforms
from matplotlib.colors import LinearSegmentedColormap
from mpl_toolkits.axes_grid1.axes_divider import make_axes_locatable

from coolbox.utilities import get_logger, GenomeRange

log = get_logger(__name__)

JuiceBoxLikeColor = LinearSegmentedColormap.from_list(
    'interaction', ['#FFFFFF', '#FFDFDF', '#FF7575', '#FF2626', '#F70000'])
JuiceBoxLikeColor.set_bad("white")
JuiceBoxLikeColor.set_under("white")
JuiceBoxLikeColor2 = LinearSegmentedColormap.from_list(
    'interaction', ['#FFFFFF', '#FFDFAF', '#FF7555', '#FF2600', '#F70000'])
JuiceBoxLikeColor2.set_bad("white")
JuiceBoxLikeColor2.set_under("white")

cmaps = {
    "JuiceBoxLike": JuiceBoxLikeColor,
    "JuiceBoxLike2": JuiceBoxLikeColor2,
}


class PlotHiCMat(object):
    JuiceBoxLikeColor = cmaps.get('JuiceBoxLike')
    JuiceBoxLikeColor2 = cmaps.get('JuiceBoxLike2')

    def __init__(self, *args, **kwargs):
        self.ax = None
        self.label_ax = None
        self.matrix = None

    def plot_matrix(self, gr: GenomeRange, gr2: GenomeRange = None):
        gr = GenomeRange(gr)

        def get_cmap(color: str):
            cm = color
            if isinstance(cm, str):
                if cm in cmaps:
                    cmap = cmaps[cm]
                else:
                    cmap = plt.get_cmap(color)
                    cmap = copy.copy(cmap)
                    lowest = cmap(0)
                    cmap.set_bad(lowest)
                    cmap.set_under(lowest)
            else:
                cmap = cm
            return cmap

        cmap = get_cmap(self.properties['color'])
        ax = self.ax
        arr = self.matrix
        c_min, c_max = self.matrix_val_range
        if gr2 is None and self.style == self.STYLE_TRIANGULAR:
            # triangular style
            scale_r = 1 / math.sqrt(2)
            r_len = gr.end - gr.start
            # Rotate image using Affine2D, reference:
            #     https://stackoverflow.com/a/50920567/8500469
            tr = transforms.Affine2D().translate(-gr.start, -gr.start) \
                .rotate_deg_around(0, 0, 45) \
                .scale(scale_r) \
                .translate(gr.start + r_len / 2, -r_len / 2)
            img = ax.matshow(arr, cmap=cmap,
                             transform=tr + ax.transData,
                             extent=(gr.start, gr.end, gr.start, gr.end),
                             aspect='auto')
        elif gr2 is None and self.style == self.STYLE_WINDOW:
            # window style
            # exist in HicMatBase
            fgr = self.fetched_gr
            scale_factor = fgr.length / gr.length
            scale_r = scale_factor / math.sqrt(2)
            length_dialog = gr.length * scale_factor
            delta_x = length_dialog * (gr.start - fgr.start) / fgr.length
            delta_x = length_dialog / 2 - delta_x
            tr = transforms.Affine2D().translate(-gr.start, -gr.start) \
                .rotate_deg_around(0, 0, 45) \
                .scale(scale_r) \
                .translate(gr.start + delta_x, -fgr.length / 2)
            img = ax.matshow(arr, cmap=cmap,
                             transform=tr + ax.transData,
                             extent=(gr.start, gr.end, gr.start, gr.end),
                             aspect='auto')
        else:
            if gr2 is None:
                gr2 = gr
            # matrix style
            img = ax.matshow(arr, cmap=cmap,
                             extent=(gr.start, gr.end, gr2.end, gr2.start),
                             aspect='auto')

        if self.norm == 'log':
            img.set_norm(colors.LogNorm(vmin=c_min, vmax=c_max))
        else:
            img.set_norm(colors.Normalize(vmin=c_min, vmax=c_max))

        return img

    def adjust_figure(self, gr: GenomeRange, gr2: GenomeRange = None):
        ax = self.ax
        if gr2 is None:
            if self.style == self.STYLE_TRIANGULAR or self.style == self.STYLE_WINDOW:

                if self.properties['depth_ratio'] == self.DEPTH_FULL:
                    depth = gr.length / 2
                else:
                    depth = (gr.length / 2) * self.properties['depth_ratio']

                if self.is_inverted:
                    ax.set_ylim(depth, 0)
                else:
                    ax.set_ylim(0, depth)
            else:
                ax.set_ylim(gr.end, gr.start)
            ax.set_xlim(gr.start, gr.end)
        else:
            ax.set_xlim(gr.start, gr.end)
            ax.set_ylim(gr2.end, gr2.start)

    def draw_colorbar(self, img):
        # plot colorbar
        if self.properties['color_bar'] == 'no':
            pass
        else:
            if hasattr(self, 'y_ax') and self.properties['color_bar'] == 'vertical':
                self.plot_colorbar(img, orientation='vertical')
            else:
                self.plot_colorbar(img, orientation='horizontal')

    def plot_colorbar(self, img, orientation='vertical'):
        if orientation == 'horizontal':
            ax_divider = make_axes_locatable(self.ax)
            if self.is_inverted:
                cax = ax_divider.append_axes("top", size=0.09, pad=0.2)
            else:
                cax = ax_divider.append_axes("bottom", size=0.09, pad=0.2)
            plt.colorbar(img, cax=cax, orientation='horizontal')
        else:  # vertical
            y_ax = self.y_ax

            if self.norm == 'log':
                from matplotlib.ticker import LogFormatter
                formatter = LogFormatter(10, labelOnlyBase=False)
                aa = np.array([1, 2, 5])
                c_min, c_max = self.matrix_val_range

                def abs_inc(num):
                    if num != 0:
                        sign = num / abs(num)
                        return int(sign * abs(num + 1))
                    else:
                        return 1

                lower_ = int(np.log10(c_min))
                upper_ = abs_inc(int(np.log10(c_max)))
                tick_values = np.concatenate([aa * 10 ** x for x in range(lower_, upper_)])

                c_bar = plt.colorbar(img, ax=y_ax, ticks=tick_values, format=formatter, fraction=0.98)
            else:
                c_bar = plt.colorbar(img, ax=y_ax, fraction=0.98)

            c_bar.solids.set_edgecolor("face")
            c_bar.ax.tick_params(labelsize='smaller')

            c_bar.ax.yaxis.set_ticks_position('left')

    def get_track_height(self, frame_width):
        """
        calculate track height dynamically.
        """
        if self.style == self.STYLE_TRIANGULAR:
            height = frame_width * 0.5
        elif self.style == self.STYLE_WINDOW:
            if 'height' in self.properties and self.properties['height'] != 'hic_auto':
                height = self.properties['height']
            else:
                height = frame_width * 0.5
        else:
            height = frame_width * 0.8

        if 'depth_ratio' in self.properties and self.properties['depth_ratio'] != self.DEPTH_FULL:
            if self.properties['style'] != self.STYLE_MATRIX:
                height = height * self.properties['depth_ratio']

        if 'color_bar' in self.properties and self.properties['color_bar'] != 'no':
            height += 1.5

        return height

    @property
    def is_inverted(self):
        if 'orientation' in self.properties and self.properties['orientation'] == 'inverted':
            return True
        else:
            # default: not inverted
            return False

    @property
    def balance(self):
        if self.properties['balance'] == 'no':
            return False
        else:
            file = self.properties['file']
            from coolbox.utilities.hic.tools import hicmat_filetype
            if hicmat_filetype(file) == '.hic':
                if self.properties['balance'] == 'yes':
                    return 'KR'  # default use KR balance
                else:
                    return self.properties['balance']
            else:
                return True

    @property
    def is_balance(self):
        return bool(self.balance)

    @property
    def style(self):
        if 'style' in self.properties:
            return self.properties['style']
        else:
            # default triangular style
            return self.STYLE_TRIANGULAR

    @property
    def matrix_val_range(self):
        small = 1e-4
        arr = self.matrix
        arr_no_nan = arr[np.isfinite(arr)]
        min_, max_ = 1e-4, 1.0

        try:
            if self.properties['min_value'] == 'auto':
                # set minimal value for color bar
                min_ = arr[arr > arr_no_nan.min()].min()
            else:
                min_ = self.properties['min_value']

            if self.properties['max_value'] == 'auto':
                max_ = arr_no_nan.max()
            else:
                max_ = self.properties['max_value']
            if max_ <= min_:
                max_ = min_ + small
        except ValueError as e:
            log.warning(e)
            return min_, max_

        return min_, max_

    @property
    def resolution(self):
        return self.properties['resolution']

    @property
    def norm(self):
        return self.properties['norm']



from typing import Iterator, Union
from .logtools import get_logger
from .filetool import opener, to_string

log = get_logger(__name__)


def to_gr(obj):
    """
    Convert object to GenomeRange.

    >>> gr = to_gr("chr1:0-1000")
    >>> (gr.chrom, gr.start, gr.end)
    ("chr1", 0, 1000)

    """
    if type(obj) is GenomeRange:
        return obj
    else:
        return GenomeRange(obj)


class GenomeRange(object):
    """
    Express a range on the genome.

    Attributes
    ----------
    chrom : str
        chromosome

    start : int
        start position

    end : int
        end position

    """

    def __init__(self, *args):
        """
        >>> range1 = GenomeRange("chr1", 1000, 2000)
        >>> str(range1)
        'chr1:1000-2000'
        >>> range2 = GenomeRange("chr2:2000-4000")
        >>> (range2.chrom, range2.start, range2.end)
        ('chr2', 2000, 4000)
        >>> range3 = GenomeRange(range2)
        >>> tuple(range3)
        >>> range4 = GenomeRange("chr1", 2000, 1000)
        Traceback (most recent call last):
        ...
        ValueError: Please check that the region end is larger than the region start. Values given: start: 2000, end: 1000
        """
        if len(args) == 1:
            if isinstance(args[0], GenomeRange):
                chrom, start, end = tuple(args[0])
            # str format
            else:
                chrom, start, end = GenomeRange.parse_region_string(args[0])

        elif len(args) == 3:
            chrom, start, end = args
        else:
            raise ValueError("inappropriate init arguments. "
                             "correct example: `range1 = GenomeRange(\"chr1:1000-2000\")` or "
                             "`range1 = GenomeRange(\"chr1\", 1000, 2000)`")

        if end < start:
            raise ValueError("Please check that the region end is larger than the region start. "
                             "Values given: start: {}, end: {}".format(start, end))

        self.chrom = chrom
        self.start = start
        self.end = end

    def __iter__(self) -> Iterator[Union[str, int]]:
        yield self.chrom
        yield self.start
        yield self.end

    @staticmethod
    def parse_region_string(region_string):
        """
        splits a region string into
        a (chrom, start, end) tuple

        Parameters
        ----------
        region_string : str
            Region string to be parsed, like: "chr:start-end"

        Return
        ------
        result : tuple of str
            Result tuple (chrom, start, end)

        >>> GenomeRange.parse_region_string("chr1:10-20")
        ('chr1', 10, 20)
        >>> GenomeRange.parse_region_string("chr1:0")
        Traceback (innermost last):
         ...
        ValueError: Failure to parse region string, please check that region format should be like "chr:start-end".
        """
        try:
            # separate the chromosome name and the location using the ':' character
            chrom, position = region_string.strip().split(":")

            # clean up the position
            for char in ",.;|!{}()":
                position = position.replace(char, '')

            position_list = position.split("-")
            region_start = int(position_list[0])
            region_end = int(position_list[1])
            return chrom, region_start, region_end

        except Exception:
            raise ValueError(f"Failure to parse region string {region_string}, please check that region format "
                             f"should be like \"chr:start-end\".")

    def change_chrom_names(self):
        """
        >>> range1 = GenomeRange("chr1", 1000, 2000)
        >>> range1.chrom
        'chr1'
        >>> range1.change_chrom_names()
        >>> range1.chrom
        '1'
        >>> range1.change_chrom_names()
        >>> range1.chrom
        'chr1'
        """
        self.chrom = change_chrom_names(self.chrom)

    @property
    def length(self):
        """
        >>> range1 = GenomeRange("chr1", 0, 1000)
        >>> range1.length
        1000
        """
        return self.end - self.start

    def __eq__(self, other):
        """
        >>> GenomeRange('chr1', 1000, 2000) == GenomeRange("chr1:1000-2000")
        True
        >>> GenomeRange("chr1:1000-2000") == GenomeRange("1:1000-2000")
        False
        """
        return str(self) == str(other)

    def __hash__(self):
        return hash(str(self))

    def __contains__(self, another):
        if another.chrom != self.chrom:
            return False
        if another.start < self.start:
            return False
        if another.end > self.end:
            return False
        return True

    def __str__(self) -> str:
        return f"{self.chrom}:{self.start}-{self.end}"

    def __repr__(self) -> str:
        return f"GenomeRange('{self}')"

def change_chrom_names(chrom):
    """
    Changes UCSC chromosome names to ensembl chromosome names
    and vice versa.

    >>> change_chrom_names("chr1")
    '1'
    >>> change_chrom_names("1")
    'chr1'
    """
    # TODO: mapping from chromosome names like mithocondria is missing
    if chrom.startswith('chr'):
        # remove the chr part from chromosome name
        chrom = chrom[3:]
    else:
        # prefix with 'chr' the chromosome name
        chrom = 'chr' + chrom

    return chrom


class GenomeLength(dict):

    def __init__(self, length_file, genome_name=""):
        self.name = genome_name
        self.length_file = length_file
        self.parse_file(length_file)

    def parse_file(self, length_file):
        with opener(length_file) as f:
            for idx, line in enumerate(f):
                line = to_string(line)
                chrom, length, *_ = line.strip().split()
                try:
                    length = int(length)
                except ValueError as detail:
                    log.warning("Error reading line #{}. The field {} is not a integer.\n"
                                "Error message: {}\n".format(idx + 1, length, detail))
                self.__setitem__(chrom, length)

    def check_range(self, genome_range):
        """
        Check a genome range is valid or not.

        (both start and end position is one based.)
        """
        if genome_range.chrom not in self:
            return False
        if genome_range.start < 1:
            return False
        if genome_range.end > self[genome_range.chrom]:
            return False
        return True

    def bound_range(self, genome_range):
        """
        Bound a genome range within reference.

        (both start and end position is one based.)
        """
        if self.check_range(genome_range):
            return genome_range
        else:
            chrom, start, end = "", 0, 0

            if genome_range.chrom not in self:
                raise ValueError("{} not in chromosome file: {}".format(
                    genome_range.chrom, self.length_file))
            else:
                chrom = genome_range.chrom

            if genome_range.start < 1:
                start = 1
            else:
                start = genome_range.start

            if genome_range.end > self[genome_range.chrom]:
                end = self[genome_range.chrom]
                if start >= end:
                    start = end - genome_range.length
                    if start < 1:
                        start = 1
            else:
                end = genome_range.end

            bounded_range = GenomeRange(chrom, start, end)
            return bounded_range


def split_genome_range(genome_range):
    """
    Little helper func.
    enforce gr is a GenomeRange object, and
    split gr to chrom, start, end
    """
    if isinstance(genome_range, str):
        genome_range = GenomeRange(genome_range)
    else:
        assert isinstance(genome_range, GenomeRange), \
            "gr is a `GenomeRange` object or a genome range str"
    chrom, start, end = genome_range.chrom, genome_range.start, genome_range.end
    return chrom, start, end



from coolbox.utilities import (
    file_to_intervaltree, GenomeRange,
    to_gr,
)
from .base import Coverage


class VlinesBase(object):
    def fetch_data(self, gr: GenomeRange):
        vlines_list = []

        if gr.chrom not in list(self.vlines_intval_tree):
            gr.change_chrom_names()

        for region in sorted(self.vlines_intval_tree[gr.chrom][gr.start - 1:gr.end + 1]):
            vlines_list.append(region.begin)
            if region.end != region.begin:
                vlines_list.append(region.end)

        return vlines_list

    def plot(self, ax, gr: GenomeRange, **kwargs):
        gr = GenomeRange(gr)
        vlines_list = self.fetch_data(gr)

        ymin, ymax = ax.get_ylim()

        ax.vlines(vlines_list, ymin, ymax,
                  linestyle=self.properties['line_style'],
                  linewidth=self.properties['line_width'],
                  color=self.properties['color'],
                  alpha=self.properties['alpha'])


class VlinesFromFile(Coverage, VlinesBase):
    """
    Vertical lines from the file.

    Parameters
    ----------
    file_ : str
        Path to file.

    color : str, optional
        Line color, default '#1e1e1e'.

    alpha : float, optional
        Line alpha value, default 0.8.

    line_style : str, optional
        Line style, default 'dashed'.

    line_width : float, optional
        Line width, default 0.5.

    name : str, optional
        The name of thr Coverage.
    """

    DEFAULT_LINE_WIDTH = 0.5
    DEFAULT_LINE_STYLE = 'dashed'
    DEFAULT_COLOR = '#1e1e1e'
    DEFAULT_ALPHA = 0.8

    def __init__(self, file_, **kwargs):
        properties_dict = {
            "file": file_,
            "color": "#1e1e1e",
            "alpha": 0.8,
            "line_style": "dashed",
            "line_width": 1,
        }
        properties_dict.update(kwargs)
        super().__init__(properties_dict)
        self.vlines_intval_tree, _, _ = file_to_intervaltree(self.properties['file'])


class Vlines(Coverage, VlinesBase):
    """
    Vertical lines.

    Parameters
    ----------
    vlines : list of {int, str}
        A list of vline positions. position can be expressed as a tuple like:
        [('chr1', 10000), ('chr2', 50000)]
        or a genome range string like:
        ['chr1:10000-10000', 'chr2:50000-50000']

    color : str, optional
        Line color, default '#1e1e1e'.

    alpha : float, optional
        Line alpha value, default 0.8.

    line_style : str, optional
        Line style, default 'dashed'.

    line_width : float, optional
        Line width, default 0.5.

    name : str, optional
        The name of thr Coverage.
    """

    def __init__(self, vlines, **kwargs):
        if not isinstance(vlines, list):
            vlines = [vlines]
        properties_dict = {
            "vlines_list": vlines,
            "color": "#1e1e1e",
            "alpha": 0.8,
            "line_style": "dashed",
            "line_width": 1,
        }
        properties_dict.update(kwargs)
        super().__init__(properties_dict)
        self.vlines_intval_tree = self.__intervaltree_from_list(self.properties['vlines_list'])

    def __intervaltree_from_list(self, vlines_list):
        from intervaltree import IntervalTree
        itree = {}
        for v in vlines_list:
            if isinstance(v, str):
                grange = GenomeRange(v)
            elif isinstance(v, tuple):
                grange = GenomeRange(v[0], v[1], v[1])
            elif isinstance(v, GenomeRange):
                grange = v
            else:
                raise ValueError("position must be a tuple or string.")
            chr_ = grange.chrom
            itree.setdefault(chr_, IntervalTree())
            itree[chr_][grange.start:grange.end + 1] = grange
        return itree



from coolbox.core.track.bed.fetch import FetchBed
from coolbox.utilities import (
    get_logger
)
from coolbox.utilities.genome import GenomeRange
from coolbox.utilities.bed import build_bed_index
from .base import BedBase

log = get_logger(__name__)


class BED(BedBase, FetchBed):
    """
    Bed Track for plotting 1d intervals data from .bed file.
    The input bed file can be bed3/bed6/bed9/bed12

    Parameters
    ----------
    file: str
        The file path of `.bed` file.


    """

    DEFAULT_PROPERTIES = {
        'labels': "on",
    }

    def __init__(self, file, **kwargs):
        properties = BED.DEFAULT_PROPERTIES.copy()
        properties.update({
            'file': file,
            **kwargs
        })
        super().__init__(**properties)
        self.bgz_file = build_bed_index(file)

    def fetch_data(self, gr: GenomeRange, **kwargs):
        intervals = self.fetch_intervals(self.bgz_file, gr)
        return intervals





from typing import Tuple

import numpy as np
import pandas as pd

from .process import ProcessHicMat
from .plot import PlotHiCMat
from ..base import Track
from coolbox.utilities.genome import GenomeRange


class HicMatBase(Track, PlotHiCMat, ProcessHicMat):
    """
    HicMatBase class for all track plotting matrix-like data.

    Parameters
    ----------
    style : {'triangular', 'window', 'matrix'}, optional
        Matrix style, default 'window'.

    cmap : str, optional
        Color map of hic matrix, default "JuiceBoxLike".

    color_bar : {'vertical', 'horizontal', 'no'}, optional
        Color bar style. default 'vertical'.

    depth_ratio : float, optional
        Depth ratio of triangular matrix, use 'full' for full depth. default 'full'.

    max_value : {float, 'auto'}, optional
        Max value of hic matrix, use 'auto' for specify max value automatically, default 'auto'.

    min_value : {float, 'auto'}, optional
        Min value of hic matrix, use 'auto' for specify min value automatically, default 'auto'.

    orientation : str, optional
        Track orientation, use 'inverted' for inverted track plot.

    resolution : {int, 'auto'}, optional
        Matrix resolution, default 'auto'.

    transform : {str, bool}, optional
        Transform for matrix, like 'log2', 'log10', default False.

    normalize : {'zscore', 'expect', 'total', False}
        Normalization method, default False.

    norm: {'log', 'no'}
        Method used when normalizing. default 'no'

    gaussian_sigma : {float, False}, optional
        Do gaussian filter(with sigma, for example 1.0) on matrix if specified. default False.

    process_func : {callable, str, False}, optional
        Process matrix with a user-defined function(receive a matrix, return a processed matrix). default False.


    """

    STYLE_TRIANGULAR = 'triangular'
    STYLE_MATRIX = 'matrix'
    STYLE_WINDOW = 'window'

    SMALL_VALUE = 1e-12
    DEPTH_FULL = 'full'

    DEFAULT_PROPERTIES = {
        "style": STYLE_WINDOW,
        # processing
        "resolution": "auto",
        "transform": 'log',
        "normalize": False,
        "norm": False,
        "gaussian_sigma": False,
        "process_func": False,
        # plotting
        'height': 'hic_auto',
        'cmap': "JuiceBoxLike",
        "color_bar": "vertical",
        "max_value": "auto",
        "min_value": "auto",
        "depth_ratio": DEPTH_FULL,
        "orientation": None,
    }

    def __init__(self, **kwargs):
        properties = HicMatBase.DEFAULT_PROPERTIES.copy()
        properties.update(kwargs)
        if 'cmap' in properties:
            properties['color'] = properties['cmap']
        super().__init__(properties)
        self.fetched_binsize = None
        self.fetched_gr = None
        self.fetched_gr2 = None

    def fetch_data(self, gr: GenomeRange, **kwargs) -> np.ndarray:
        """
        Fetch the raw matrix should be plotted. Normally it's a matrix with raw contacts

        Parameters
        ----------
        gr2 : GenomeRange, optional, keyword argument

        Returns
        -------
        matrix : np.array
            Hi-C raw contact matrix.

        """
        raise NotImplementedError

    def fetch_plot_data(self, gr: GenomeRange, **kwargs) -> np.ndarray:
        """
        Fetch 2d contact matrix, the matrix may be processed in case
        'transform', 'normalize', 'gaussian_sigma', 'process_func' exits in properties.

        Parameters
        ----------
        gr2 : GenomeRange, optional

        gr_updated: bool, optional
            If the input GenomeRange has been updated. default False
            Default False means that the input gr will be expanded in window mode

        Returns
        -------
        matrix : np.array
            Processed hic matrix used for plotting.
        """
        gr2 = kwargs.get('gr2')
        if self.properties['style'] == self.STYLE_WINDOW and not kwargs.get("gr_updated", False):
            gr, gr2 = self.fetch_window_genome_range(gr, gr2)
            kwargs.update({'gr2': gr2})
        arr = self.fetch_data(gr, **kwargs)
        # store fetched_gr
        self.fetched_gr = gr
        self.fetched_gr2 = gr2
        return self.process_matrix(arr)

    def plot(self, ax, gr: GenomeRange, **kwargs):
        """
        Plot matrix

        Parameters
        ----------
        ax
        gr2 : GenomeRange, optional
        """
        self.ax = ax
        # fetch processed plot_data
        self.matrix = self.fetch_plot_data(gr, **kwargs)
        # plot matrix
        img = self.plot_matrix(gr, kwargs.get('gr2'))
        self.adjust_figure(gr, kwargs.get('gr2'))
        self.draw_colorbar(img)
        self.plot_label()

    def fetch_pixels(self, gr: GenomeRange, **kwargs) -> pd.DataFrame:
        raise NotImplementedError

    def infer_binsize(self, gr: GenomeRange, **kwargs) -> int:
        raise NotImplementedError

    def fetch_window_genome_range(self, gr: GenomeRange, gr2: GenomeRange = None) -> Tuple[GenomeRange, GenomeRange]:
        from copy import copy
        fetch_gr = copy(gr)
        dr = self.properties['depth_ratio']
        dr = 1.0 if dr == "full" else dr
        dr = min(1.0, dr + 0.05)
        x = int(gr.length * dr // 2)
        fetch_gr.start = gr.start - x
        fetch_gr.end = gr.end + x

        if fetch_gr.start < 0:
            fetch_gr.start = 0

        return fetch_gr, gr2



from typing import Union

from collections import OrderedDict

import svgutils.compose as sc
import matplotlib.pyplot as plt

from coolbox.utilities.figtools import cm2inch
from coolbox.utilities.filetool import get_uniq_tmp_file
from coolbox.utilities import GenomeRange
from coolbox.core.track import Track
from coolbox.core.frame import Frame
from .base import SuperFrame


class JointView(SuperFrame):
    """Compose two track and a center matrix.

    https://github.com/GangCaoLab/CoolBox/issues/12

    Parameters
    ----------
    center : Track
        Center track for show 'contact map-like' plot,
        should support for plotting images with 2d GenomeRanges.

    top : Frame, optional
        Frame plot in the top of the center track.

    right : Frame, optional

    bottom : Frame, optional

    left : Frame, optional

    center_width : float
        The width of center track, unit in cm. default 20.

    trbl : str
        sub-frames(top right bottom left) use which genome range(first or second),
        default '1212', which means: top -> 1, right -> 2, bottom -> 1, left -> 2

    space : float
        Space between frame and center, unit in cm. default 0.5
    """

    def __init__(self,
                 center: Track,
                 top: Union[Frame, Track] = None,
                 right: Union[Frame, Track] = None,
                 bottom: Union[Frame, Track] = None,
                 left: Union[Frame, Track] = None,
                 **kwargs,
                 ):
        sub_frames = OrderedDict({
            "top": top,
            "right": right,
            "bottom": bottom,
            "left": left,
        })
        sub_frames = {k: v for k, v in sub_frames.items() if v is not None}
        self.__check_sub_frames(center, sub_frames)
        # explicitely use matrix style
        center.properties['style'] = 'matrix'
        properties = {
            "sub_frames": sub_frames,
            "center_track": center,
            "center_width": 20,
            "trbl": "1212",
            "space": 1,
            "cm2px": 28.5,
            "padding_left": 1,
        }
        properties.update(**kwargs)

        super().__init__(properties)
        self.__adjust_sub_frames_width()

    def cm2px(self, vec):
        return [i * self.properties['cm2px'] for i in vec]

    @staticmethod
    def __check_sub_frames(center, sub_frames):
        from ..frame import Frame
        from ...track.base import Track
        sub_f_names = ", ".join(sub_frames.keys())
        if (not isinstance(center, Track)) and (not hasattr(center, "plot_joint")):
            raise TypeError("center should be a Track type instance with plot_joint method, "
                            "for example Cool, DotHiC, ...")
        for k, f in sub_frames.items():
            if not isinstance(f, Frame):
                if isinstance(f, Track):
                    sub_frames[k] = (Frame() + f)  # convert track to frame
                else:
                    raise TypeError(f"{sub_f_names} should be Frame object.")

    def __adjust_sub_frames_width(self):
        for k, f in self.properties['sub_frames'].items():
            width_ratios = f.properties['width_ratios']
            middle_ratio = width_ratios[1]
            new_width = self.properties['center_width'] / middle_ratio
            f.properties['width'] = new_width

    def plot_center(self, gr1: GenomeRange, gr2: GenomeRange):
        center_track = self.properties['center_track']
        size = cm2inch(self.properties['center_width'])
        fig, ax = plt.subplots(figsize=(size, size))
        center_track.plot(ax, gr2, gr2=gr1)
        center_track.plot_coverages(ax, gr2, gr1)
        ax.set_axis_off()
        path = get_uniq_tmp_file(prefix='center', suffix='.svg')
        fig.subplots_adjust(wspace=0, hspace=0.0, left=0, right=1, bottom=0, top=1)
        fig.savefig(path)
        plt.close()
        return sc.SVG(path)

    def frame_granges(self, gr1=None, gr2=None):
        self.goto(gr1, gr2)
        gr1, gr2 = self.current_range

        trbl = self.properties['trbl']

        orientations = ['top', 'right', 'bottom', 'left']
        frame2grange = {
            k: (gr1, gr2) if (trbl[orientations.index(k)] == '1') else (gr2, gr1)
            for k in orientations
        }

        return frame2grange

    def plot(self, gr1=None, gr2=None):
        """

        Parameters
        ----------
        gr1 : {str, GenomeRange}
            First genome range

        gr2 : {str, GenomeRange}, optional
            Second genome range
        """
        frame2grange = self.frame_granges(gr1, gr2)
        gr1, gr2 = self.current_range
        sub_frames = self.properties['sub_frames']

        frame_svgs = self.plot_frames(frame2grange)
        center_svg = self.plot_center(gr1, gr2)

        center_offsets = self.__get_center_offsets(sub_frames)

        center_svg.move(*self.cm2px(center_offsets))
        self.__transform_sub_svgs(frame_svgs, sub_frames, center_offsets)

        figsize = self.cm2px(self.__get_figsize(sub_frames))
        fig = sc.Figure(f"{figsize[0]}px", f"{figsize[1]}px",
                        sc.Panel(center_svg),
                        *[sc.Panel(svg) for svg in frame_svgs.values()])
        return fig

    def fetch_data(self, gr1=None, gr2=None) -> dict:
        """

        Parameters
        ----------
        gr1 : {str, GenomeRange}, optional
            First genome range

        gr2 : {str, GenomeRange}, optional
            Second genome range
        """
        tracks_data = {}
        frame2grange = self.frame_granges(gr1, gr2)
        gr1, gr2 = self.current_range
        sub_frames = self.properties['sub_frames']

        for pos, fr in sub_frames.items():
            tracks_data[pos] = sub_frames[pos].fetch_data(frame2grange[pos][0], gr2=frame2grange[pos][1])

        tracks_data['center'] = self.properties['center_track'].fetch_data(gr1, gr2=gr2)

        return tracks_data

    def __transform_sub_svgs(self, sub_svgs, sub_frames, center_offsets):
        c_width = self.properties['center_width']
        space = self.properties['space']
        pd_left = self.properties['padding_left']
        if 'top' in sub_svgs:
            s = sub_svgs['top']
            offsets = [pd_left, 0]
            if 'left' in sub_svgs:
                offsets[0] += sub_frames['left'].properties['height'] + space
            s.move(*self.cm2px(offsets))
        if 'right' in sub_svgs:
            f = sub_frames['right']
            s = sub_svgs['right']
            s.rotate(90)
            wr = f.properties['width_ratios']
            right_offsets = [
                center_offsets[0] + c_width + f.properties['height'] + space,
                center_offsets[1] - f.properties['width'] * wr[0]
            ]
            right_offsets = self.cm2px(right_offsets)
            s.move(*right_offsets)
        if 'bottom' in sub_svgs:
            s = sub_svgs['bottom']
            offsets = [pd_left, c_width]
            if 'left' in sub_svgs:
                offsets[0] += sub_frames['left'].properties['height'] + space
            if 'top' in sub_svgs:
                offsets[1] += sub_frames['top'].properties['height']
            offsets = self.cm2px(offsets)
            s.move(*offsets)
        if 'left' in sub_svgs:
            s = sub_svgs['left']
            f = sub_frames['left']
            s.rotate(90)
            offsets = [pd_left + f.properties['height'], 0]
            if 'top' in sub_svgs:
                offsets[1] += sub_frames['top'].properties['height']
            offsets = self.cm2px(offsets)
            s.move(*offsets)

    def __get_center_offsets(self, sub_frames):
        space = self.properties['space']
        center_offsets = [self.properties['padding_left'], 0]  # x, y (left, top)
        if 'top' in sub_frames:
            f = sub_frames['top']
            center_offsets[0] += f.properties['width'] * f.properties['width_ratios'][0]
            center_offsets[1] += space + f.properties['height']
        if 'bottom' in sub_frames:
            f = sub_frames['bottom']
            if 'top' not in sub_frames:
                center_offsets[0] += f.properties['width'] * f.properties['width_ratios'][0]
        if 'left' in sub_frames:
            center_offsets[0] += space + sub_frames['left'].properties['height']
        return center_offsets

    def __get_figsize(self, sub_frames):
        space = self.properties['space']
        center_width = self.properties['center_width']
        size = [center_width, center_width]  # width, height
        if 'top' in sub_frames:
            f = sub_frames['top']
            size[0] = f.properties['width']
            size[1] += f.properties['height'] + space
        if 'bottom' in sub_frames:
            f = sub_frames['bottom']
            size[0] = f.properties['width']
            size[1] += f.properties['height'] + space
        if 'left' in sub_frames:
            f = sub_frames['left']
            size[0] += f.properties['height'] + space
        if 'right' in sub_frames:
            f = sub_frames['right']
            size[0] += f.properties['height'] + space
        self.properties['width'] = size[0]
        self.properties['height'] = size[1]
        return size

    def add_track(self, track, pos=None):
        sub_frames = self.properties['sub_frames']
        if pos is None:
            pos = list(sub_frames.keys())[-1]
        frame = sub_frames[pos]
        frame.add_track(track)



import abc
from collections import OrderedDict

import svgutils.compose as sc
from IPython.display import SVG

from coolbox.utilities.filetool import get_uniq_tmp_file
from coolbox.utilities import GenomeRange
from ..base import FrameBase


class SuperFrame(FrameBase, abc.ABC):
    """Bigger frame composed by normal frames,
    compose figure using svgutils,
    this allow compose big figures which the matplotlib can not do.
    """

    def __init__(self, properties_dict, **kwargs):
        assert "sub_frames" in properties_dict
        super().__init__(properties_dict, **kwargs)
        self.current_range = [None, None]

    def plot_frames(self, frame2grange):
        """Plot each frame by a given GenomeRange object."""
        res = OrderedDict()
        for k, f in self.properties['sub_frames'].items():
            gr, gr2 = frame2grange[k]
            path = get_uniq_tmp_file(prefix="frame_", suffix=".svg")
            fig = f.plot(gr, gr2)
            fig.subplots_adjust(wspace=0, hspace=0.0, left=0, right=1, bottom=0, top=1)
            fig.savefig(path)
            svg = sc.SVG(path)
            res[k] = svg
        return res

    @property
    def tracks(self):
        sub_frames = self.properties['sub_frames']
        tracks = OrderedDict()
        for f in sub_frames.values():
            tracks.update(f.tracks)
        return tracks

    def goto(self, gr1=None, gr2=None):
        if gr1 is not None:
            gr1 = GenomeRange(gr1)
        if gr2 is not None:
            gr2 = GenomeRange(gr2)
        if gr1 is None:
            gr1 = self.current_range[0]
        if gr2 is None:
            gr2 = gr1

        if gr1 is None or gr2 is None:
            raise ValueError("No history gr found.")
        self.current_range = [gr1, gr2]

    def plot(self, gr1=None, gr2=None):
        """

        Parameters
        ----------
        gr1 : {str, GenomeRange}
            First genome range

        gr2 : {str, GenomeRange}, optional
            Second genome range
        """
        raise NotImplementedError

    def show(self):
        gr1, gr2 = self.current_range
        svg = self.plot(gr1, gr2=gr2)
        path = get_uniq_tmp_file(prefix="sub_frame", suffix=".svg")
        svg.save(path)
        return SVG(path)



"""
CoolBox API

example:

``` Python
from coolbox.api import *

bsr = Browser()

# open files
bsr.open("GM12878-rna-1.bw")
bsr.open("GM12878-chip-h3k27ac-1.bw")
bsr.open("GM12878-Hind3-1.cool")

bsr.show()

```

Or, using another style(ggplot like) API.

``` Python
from coolbox.api import *

frame = XAxis() + \
        BigWig("GM12878-rna-1.bw) + Color("#66ccff") + \
        BigWig("GM12878-chip-h3k27ac-1.bw") + Color("#ff9c9c") + \
        Cool("GM12878-Hind3-1.cool") + ColorMap("#0000ff", "#ff0000")

bsr = Browser(frame)

bsr.show()

```


Element types:

* Track
* Coverage
* Frame
* Feature
    - FrameFeature
* WidgetsPanel
* Browser


The rule of element composition:
```
    Track + Track = Frame
    Track + Feature = Track
    Track + Coverage = Track
    Frame + Track = Frame
    Frame + Coverage = Frame
    Frame + Feature = Frame
    Frame + FrameFeature = Frame
    Frame + Frame = Frame
    Frame + WidgetsPanel = Browser
    Coverage + Feature = Coverage

    Frame * Feature = Frame
    Frame * Coverage = Frame
```
"""

import warnings

warnings.filterwarnings('ignore')

from coolbox.core.track import *
from coolbox.core.frame import *
from coolbox.core.feature import *
from coolbox.core.coverage import *
from coolbox.core.browser import *
from coolbox.core.frame.superframe import JointView
from coolbox.utilities.genome import GenomeRange



from typing import Union, Callable

from scipy import sparse
import numpy as np

from coolbox.utilities import GenomeRange, get_logger

from ..hicmat import HicMatBase, HiCMat
from .base import HistBase

log = get_logger(__name__)


class HicFeature(HistBase):
    """
    HicFeature base track

    Parameters
    -----------
    hicmat: {str, HicMatBase}
        The input hicmat file or HicMatBase object used to calculate di score.

    args_hic : dict, optional
        Key words arguments used to create Cool/DotHic tack when the input hicmat is a file path

    """
    IGNORE_DIAGS = 3

    def __init__(self, hicmat: Union[str, HicMatBase], args_hic=None, **kwargs):
        self.hicmat = HiCMat(hicmat, **(args_hic or {}))
        properties = {
            'file': self.hicmat.properties.get('file', ""),
        }
        properties.update(kwargs)
        super().__init__(**properties)


class DiScore(HicFeature):
    """
    Directionnality index track.

    Parameters
    ----------
    hicmat: {str, HicMatBase}
        The input hicmat file or HicMatBase object used to calculate di score.

    window_size: int, optional
        Width of the diamond region along the matrix used to calculate di. default: 40

    method: {'standard', 'adaptive'}, optional
        Method used for calculating di. default: 'adaptive'

    args_hic : dict, optional
        Key words arguments used to create Cool/DotHic tack when the input hicmat is a file path
    """

    DEFAULT_PROPERTIES = {
        'style': HistBase.STYLE_FILL
    }

    def __init__(self,
                 hicmat: Union[str, HicMatBase],
                 window_size: int = 40,
                 method="adaptive",
                 args_hic: dict = None,
                 **kwargs):
        properties = DiScore.DEFAULT_PROPERTIES.copy()
        properties.update(kwargs)
        super().__init__(hicmat, args_hic, **properties)
        self.window_size = window_size
        self.method = method

    def fetch_data(self, gr: GenomeRange, **kwargs) -> np.ndarray:
        ds_mat: np.ndarray = self.hicmat.fetch_data(gr)
        mlen = ds_mat.shape[0]
        if mlen < self.window_size * 2:
            return np.zeros(mlen)
        mat = sparse.csr_matrix(ds_mat)
        window_size = self.window_size
        ignore_diags = self.IGNORE_DIAGS

        x, y = mat.nonzero()
        dis = y - x
        # filter for non_nan pixels in valid region
        max_len = ignore_diags + window_size
        available = ((dis >= ignore_diags)
                     & (dis < max_len)
                     & (x >= max_len - 1)
                     & (y <= mlen - max_len))
        x, y = x[available], y[available]
        values = np.array(mat[x, y]).ravel()
        non_nan = ~np.isnan(values)
        x, y, values = x[non_nan], y[non_nan], values[non_nan]
        dis = y - x

        contacts_up = np.zeros((mlen, window_size), dtype=mat.dtype)
        contacts_down = np.zeros((mlen, window_size), dtype=mat.dtype)
        for shift in range(ignore_diags, max_len):
            window_pos = shift - ignore_diags
            mask = dis == shift
            _x, _values = x[mask], values[mask]
            contacts_up[_x + shift, window_pos] = _values
            contacts_down[_x, window_pos] = _values
        contacts_up[:max_len, :] = 0
        contacts_down[:max_len, :] = 0

        return self.di_methods(self.method)(contacts_up, contacts_down)

    @classmethod
    def di_methods(cls, method: str) -> Callable[[np.ndarray, np.ndarray], np.ndarray]:
        def standard(up, down):
            """
            Compute directionality index described in:\n
            Jesse R.Dixon 2012. Topological domains in mammalian genomes identified by analysis of chromatin interactions.
            """
            up = up.sum(axis=1)
            down = down.sum(axis=1)
            expected = (up + down) / 2.0
            di_array = (np.sign(down - up) *
                        ((up - expected) ** 2 + (down - expected) ** 2)
                        / expected)

            return di_array

        def adaptive(up, down):
            """
            Compute directionality index described in:\n
            Xiao-Tao Wang 2017.HiTAD: Detecting the structural and functional hierarchies of topologically associating
            domains from chromatin interactions.
            """
            window_size = up.shape[1]
            mean_up = up.mean(axis=1)
            mean_down = down.mean(axis=1)
            var_up = np.square(up - mean_up[:, None]).sum(axis=1)
            var_down = np.square(down - mean_down[:, None]).sum(axis=1)
            denom = np.sqrt((var_up + var_down) / (window_size * (window_size - 1)))
            denom[denom == 0] = 1
            di_array = (mean_down - mean_up) / denom

            return di_array

        if method not in locals():
            raise RuntimeError("Only support methods {}".format(list(locals().keys())))

        return locals()[method]


class InsuScore(HicFeature):
    """
    Insulation score track.

    Parameters
    ----------
    hicmat: {str, HicMatBase}
        The input hicmat file or HicMatBase object used to calculate di score.

    window_size: {int, str}, optional
        Width of the diamond region along the matrix used to calculate di. default: 20
        window_size can also be a str with format like "20-40" representing a range of window_sizes.

    normalize: bool, optional
        Weather to log-nomalize the insulation score array. default: true

    method: {'standard', 'adaptive'}, optional
        Method used for calculating di. default: 'adaptive'

    args_hic : dict, optional
        Key words arguments used to create Cool/DotHic tack when the input hicmat is a file path
    """

    DEFAULT_PROPERTIES = {
        "style": "line",
        "cmap": "coolwarm_r",
    }

    def __init__(self,
                 hicmat: Union[str, HicMatBase],
                 window_size: Union[int, str] = 20,
                 normalize: bool = True,
                 args_hic: dict = None,
                 **kwargs):
        properties = InsuScore.DEFAULT_PROPERTIES.copy()
        properties.update(kwargs)
        super().__init__(hicmat, args_hic, **properties)
        self.window_size = window_size
        self.normalize = normalize

    def fetch_data(self, gr: GenomeRange, **kwargs) -> np.ndarray:
        window_size = self.window_size
        try:
            if isinstance(window_size, int):
                window_size = [window_size]
            elif isinstance(window_size, str):
                st, ed = window_size.split("-")
                window_size = list(range(int(st), int(ed)))
            else:
                raise ValueError
        except Exception:
            raise ValueError("window_size should be like int: 12 or str: '10-50'.")

        insus = []
        # for multiple window sizes
        for ws in window_size:
            mat: np.ndarray = self.hicmat.fetch_data(gr)
            mlen = mat.shape[0]
            if mlen < ws * 2:
                insus.append(np.zeros(mlen))
            else:
                insu = np.full(mlen, np.nan, dtype=mat.dtype)
                for row in range(ws, mlen - ws):
                    sub_mat = mat[row - ws: row, row + 1: row + ws + 1]
                    insu[row] = np.nanmean(sub_mat)

                if self.normalize:
                    insu = np.log2(insu / np.nanmean(insu))
                    insu[~np.isfinite(insu)] = 0
                insus.append(insu)

        return insus[0] if len(insus) == 1 else np.array(insus)[::-1]


class Virtual4C(HicFeature):
    """
    Track for view virtual 4C related to a certain genome position,
    and a HiC Track (include `Cool` and `DotHiC`).

    Parameters
    ----------
    hicmat : {str, `Cool`, `DotHiC`}
        Related hic track or Hi-C file path.

    genome_position : str
        related genome position, like: 'chr1:2000000-2000000'

    bin_width : int, optional
        How many bin used for calculate the mean value.
        default 3

    args_hic : dict, optional
        Argument for create hic track, needed only if first argument is a path.
    """

    DEFAULT_PROPERTIES = {
        "style": HistBase.STYLE_LINE,
        "line_width": 1,
        "color": "#2855d8",
        "bin_width": 3,
    }

    def __init__(self,
                 hicmat: Union[str, HicMatBase],
                 genome_position: str,
                 args_hic: dict = None,
                 **kwargs):
        properties = Virtual4C.DEFAULT_PROPERTIES.copy()
        properties.update({
            "genome_position": genome_position,
            **kwargs,
        })
        super().__init__(hicmat, args_hic, **properties)
        self.position = GenomeRange(self.properties['genome_position'])
        self.bin_width = self.properties['bin_width']

    def fetch_data(self, gr: GenomeRange, **kwargs) -> np.ndarray:
        # fetch mean array
        from copy import copy
        bin_width = self.bin_width
        position = self.position
        binsize = self.hicmat.infer_binsize(gr)
        window_range = copy(position)
        offset_ = (bin_width - 1) // 2
        assert offset_ >= 0, "bin width must >= 1"
        window_range.start = window_range.start - offset_ * binsize
        window_range.end = window_range.end + offset_ * binsize
        arr = self.hicmat.fetch_data(window_range, gr2=gr)
        mean_arr = np.nanmean(arr, axis=0)
        return mean_arr



from coolbox.utilities import GenomeRange
from .base import Track


# TODO change properties to new mode
class Spacer(Track):
    """
    The spacer track
    Does not have any real content, just used to split two tracks.

    Parameters
    ----------
    height : float, optional
        The height of Spacer track. (Default: Spacer.DEFAULT_HEIGHT)

    """

    DEFAULT_HEIGHT = 1

    def __init__(self, *args, **kwargs):
        properties_dict = {
            'height': Spacer.DEFAULT_HEIGHT,
        }
        if args:
            height = args[0]
            properties_dict['height'] = height
        properties_dict.update(kwargs)

        super().__init__(properties_dict)

    def fetch_data(self, gr: GenomeRange, **kwargs):
        pass

    def plot(self, ax, gr: GenomeRange, **kwargs):
        self.ax = ax
        ax.set_xlim(gr.start, gr.end)


class HLine(Track):
    """
    Horizontal line track.
    Used for add a horizontal line between two tracks.

    Parameters
    ----------
    line_width : float, optional
        (Default: HLine.DEFAULT_LINE_WIDTH)

    line_style : str, optional
        (Default: HLine.DEFAULT_LINE_STYLE)

    color : str, optional
        (Default: HLine.DEFAULT_COLOR)

    alpha : float, optional
        (Default: HLine.DEFAULT_ALPHA)

    height : float, optional
        The height of Spacer track. (Default: HLine.DEFAULT_HEIGHT)

    name : str, optional
        Track's name.
    """

    DEFAULT_PROPERTIES = {
        "line_width": 1.0,
        "line_style": '--',
        "height": 0.05,  # this just a empiric value,
        "color": '#000000',
        'alpha': 0.75
    }

    def __init__(self, **kwargs):
        properties = HLine.DEFAULT_PROPERTIES.copy()
        properties.update(kwargs)
        super().__init__(properties)

    def fetch_data(self, gr: GenomeRange, **kwargs):
        pass

    def plot(self, ax, gr: GenomeRange, **kwargs):
        self.ax = ax

        ax.set_xlim(gr.start, gr.end)
        ax.hlines(0, gr.start, gr.end,
                  linestyles=self.properties['line_style'],
                  linewidth=self.properties['line_width'],
                  colors=self.properties['color'],
                  alpha=self.properties['alpha'])


class XAxis(Track):
    """
    The x axis track.

    Parameters
    ----------
    height : float, optional
        Height of Spacer track. (Default: XAxis.DEFAULT_HEIGHT)

    fontsize : int, optional
        Font size of XAxis. (Default: XAxis.DEFAULT_FONTSIZE)

    where : {'top', 'bottom'}, optional
        The position of tick labels relative to the axis.
        (Default: 'bottom')

    name (str, optional):
        Track's name.
    """

    DEFAULT_PROPERTIES = {
        "where": "bottom",
        "height": 2,
        "fontsize": 15,
    }

    def __init__(self, **kwargs):
        properties = XAxis.DEFAULT_PROPERTIES.copy()
        properties.update(kwargs)
        super().__init__(properties)

    def fetch_data(self, gr: GenomeRange, **kwargs):
        pass

    def plot(self, ax, gr: GenomeRange, **kwargs):
        self.ax = ax

        ax.set_xlim(gr.start, gr.end)
        ticks = ax.get_xticks()
        if ticks[-1] - ticks[1] <= 1e5:
            labels = ["{:,.0f}".format((x / 1e3))
                      for x in ticks]
            labels[-2] += " Kb"

        elif 1e5 < ticks[-1] - ticks[1] < 4e6:
            labels = ["{:,.0f}".format((x / 1e3))
                      for x in ticks]
            labels[-2] += " Kb"
        else:
            labels = ["{:,.1f} ".format((x / 1e6))
                      for x in ticks]
            labels[-2] += " Mbp"

        ax.axis["x"] = ax.new_floating_axis(0, 0.5)

        ax.axis["x"].axis.set_ticklabels(labels)
        ax.axis['x'].axis.set_tick_params(which='minor', bottom='on')

        ax.axis["x"].major_ticklabels.set(size=int(self.properties['fontsize']))

        if 'where' in self.properties and self.properties['where'] == 'top':
            ax.axis["x"].set_axis_direction("top")


class ChromName(Track):
    """
    Track for show chromosome name.

    Parameters
    ----------
    fontsize : float
        Font name to show.

    offset : float
        Offset ratio to the start position.
    """

    DEFAULT_PROPERTIES = {
        "fontsize": 50,
        "offset": 0.45,
    }

    def __init__(self, **kwargs):
        properties = ChromName.DEFAULT_PROPERTIES.copy()
        properties.update(kwargs)
        super().__init__(properties)

    def fetch_data(self, gr: GenomeRange, **kwargs):
        return gr.chrom  # return chromosome name

    def plot(self, ax, gr: GenomeRange, **kwargs):
        x = gr.start + self.properties['offset'] * (gr.end - gr.start)
        ax.text(x, 0, gr.chrom, fontsize=self.properties['fontsize'])
        ax.set_xlim(gr.start, gr.end)



from copy import copy

from coolbox.utilities import (
    op_err_msg, get_coverage_stack, get_feature_stack,
    format_properties, GenomeRange
)


class Coverage(object):
    """
    Coverage base class.

    `Coverage` is the plots at the upper layer of Track.

    >>> from coolbox.core.track.base import Track
    >>> c1 = Coverage({})
    >>> c1.properties
    {'name': 'Coverage.1'}
    >>> c2 = Coverage({})
    >>> t1 = Track({})
    >>> t2 = c1 + c2 + t1
    >>> len(t2.coverages)
    2
    >>> assert type(c1 + c2) is CoverageStack
    >>> t3 = Track({})
    >>> frame = t2 + t3
    >>> frame = frame + Coverage({})
    >>> len(list(frame.tracks.values())[-1].coverages)
    1
    """

    def __new__(cls, *args, **kwargs):
        if hasattr(cls, "_counts"):
            cls._counts += 1
        else:
            cls._counts = 1
        return super().__new__(cls)

    def __init__(self, properties_dict):
        self.properties = format_properties(properties_dict)
        name = self.properties.get("name")
        if name is not None:
            assert isinstance(name, str), "Coverage name must be a `str`."
        else:
            name = self.__class__.__name__ + ".{}".format(self.__class__._counts)
        self.properties['name'] = name

        stack = get_feature_stack()
        for feature in stack:
            self.properties.update(feature.properties)

        self.track = None

    @property
    def name(self):
        return self.properties['name']

    @name.setter
    def name(self, value):
        self.properties['name'] = value

    def __add__(self, other):
        from ..track.base import Track
        from ..frame.base import FrameBase
        from ..feature import Feature

        if isinstance(other, Track):
            result = copy(other)
            result.append_coverage(self)
            return result
        elif isinstance(other, FrameBase):
            result = copy(other)
            if len(result.tracks) > 1:
                first = list(result.tracks.values())[0]
                first.append_coverage(self, pos='bottom')
            return result
        elif isinstance(other, Feature):
            result = copy(self)
            result.properties.update(other.properties)
            return result
        elif isinstance(other, Coverage):
            stack = CoverageStack([self, other])
            return stack
        elif isinstance(other, CoverageStack):
            result = copy(other)
            result.to_bottom(self)
            return result
        else:
            raise TypeError(op_err_msg(self, other))

    def __mul__(self, other):
        from ..frame.base import FrameBase
        if isinstance(other, FrameBase):
            result = copy(other)
            result.add_cov_to_tracks(self)
            return result
        else:
            raise TypeError(op_err_msg(self, other, op='*'))

    def __enter__(self):
        stack = get_coverage_stack()
        stack.append(self)
        return self

    def __exit__(self, type, value, traceback):
        stack = get_coverage_stack()
        stack.pop()

    def check_track_type(self, allow):
        valid = any([isinstance(self.track, type_) for type_ in allow])
        if not valid:
            msg = "{} coverage's track must be a instance of {}".format(
                self.track.__class__.__name__,
                [type_.__name__ for type_ in allow]
            )
            raise ValueError(msg)


class CoverageStack(object):
    """
    Denote a stack of Coverage.

    ps: this "Stack" is actually a "Deque",
        name it "Stack" is just for imaging it vertically.

    Parameters
    ----------
    coverages : list of coolbox.core.coverage.Coverage
        coverages list.
    """

    def __init__(self, coverages):
        self.coverages = coverages

    def to_top(self, cov):
        self.coverages.append(cov)

    def to_bottom(self, cov):
        self.coverages.insert(0, cov)

    def __add__(self, other):
        from ..track.base import Track
        from ..frame.base import FrameBase
        from ..feature import Feature

        if isinstance(other, Coverage):
            result = copy(self)
            result.to_top(other)
            return result
        elif isinstance(other, Track):
            result = copy(other)
            result.pile_coverages(self.coverages, pos='bottom')
            return result
        elif isinstance(other, FrameBase):
            result = copy(other)
            if len(result.tracks) != 0:
                first = list(result.tracks.values())[0]
                first.pile_coverages(self.coverages, pos='bottom')
            return result
        elif isinstance(other, Feature):
            result = copy(self)
            if len(result.coverages) != 0:
                last = result.coverages[-1]
                last.properties.update(other.properties)
            return result
        else:
            raise TypeError(op_err_msg(self, other))


def has_data_range(track_class):
    from coolbox.core.track import HistBase, Hist, ArcsBase, Arcs
    if isinstance(track_class, type):
        return HistBase in track_class.__mro__ or ArcsBase in track_class.__mro__
    else:
        return track_class in (Hist, Arcs)


def track_to_coverage(track_class):
    def init(self, *args, **kwargs):
        if has_data_range(track_class):
            kwargs.update({
                "show_data_range": False,
            })
        self.track_instance = track_class(*args, **kwargs)
        self.properties = self.track_instance.properties

    def plot(self, ax, gr: GenomeRange, **kwargs):
        if hasattr(self, 'track'):
            if has_data_range(track_class):
                # update height when plot
                self.track_instance.properties['height'] = self.track.properties['height']
        self.track_instance.plot(ax, gr, **kwargs)

    # TODO , other track methods
    cov_class = type(
        track_class.__name__ + "Coverage",
        (Coverage,),
        {
            "__init__": init,
            "__doc__": track_class.__doc__,
            "plot": plot
        }
    )

    return cov_class



from .base import HistBase
from .bedgraph import BedGraph
from .bigwig import BigWig, ABCompartment
from .hicfeature import DiScore, InsuScore, Virtual4C
from .bam import BAMCov
from .snp import SNP


def Hist(file, *args, **kwargs) -> HistBase:
    """Compose BigWig, BedGraph, SNP, BAMCov track automatically based on type of file
    extension.(.bw, .bigwig, .snp, .vcf, .bedgraph, .bam, .sam)
    """
    if file.endswith((".bw", ".bigwig", ".bigWig")):
        return BigWig(file, *args, **kwargs)
    elif file.endswith(('.bedgraph', ".bg", ".bedgraph.bgz", ".bg.bgz")):
        return BedGraph(file, *args, **kwargs)
    elif file.endswith((".snp", ".vcf")):
        return SNP(file, *args, **kwargs)
    elif file.endswith(".bam", ".sam"):
        return BAMCov(file, *args, **kwargs)
    else:
        raise NotImplementedError("Hist only support .bigwig or .bedgraph file now.")



from ..genome import GenomeRange


def hicmat_filetype(path):
    if path.endswith(".hic"):
        return '.hic'
    else:
        p = path.split("::")[0]
        if p.endswith((".cool", ".mcool")):
            return '.cool'
        else:
            raise NotImplementedError("File type of {} not supported for HicMat".format(path))


def infer_resolution(genome_range: GenomeRange, resolutions: object, bin_thresh: object = 1000) -> object:
    """
    Inference appropriate resolution.
    """
    resolutions.sort()
    reso = resolutions[0]
    for r in resolutions:
        num_bins = genome_range.length // reso
        if num_bins >= bin_thresh:
            reso = r
        else:
            break
    return reso


def is_multi_cool(cooler_file):
    """
    Judge a cooler is muliti-resolution cool or not.

    Parameters
    ----------
    cooler_file : str
        Path to cooler file.
    """
    import re
    if re.match(".+::.+$", cooler_file):
        return False

    import h5py
    h5_file = h5py.File(cooler_file, 'r')
    is_multi = 'pixels' not in h5_file  # use "pixels" group distinguish is multi-cool or not
    h5_file.close()
    return is_multi


def get_cooler_resolutions(cooler_file, is_multi=True):
    """
    Get the resolutions of a muliti-cooler file

    Parameters
    ----------
    cooler_file : str
        Path to cooler file.
    """
    import h5py
    h5_file = h5py.File(cooler_file, 'r')
    if is_multi:
        if 'resolutions' in h5_file:
            resolutions = list(h5_file['resolutions'])
            resolutions = [int(res) for res in resolutions]
        else:
            resolutions = [int(h5_file[i].attrs['bin-size']) for i in list(h5_file)]
        resolutions.sort()
        h5_file.close()
    else:
        resolutions = [int(h5_file.attrs['bin-size'])]
    return resolutions



from typing import Union, Tuple

import numpy as np
import pandas as pd

from coolbox.utilities import GenomeRange, get_logger
from ..base import Track
from .plot import PlotHist

log = get_logger(__name__)

HistData = Union[Tuple[np.ndarray, np.ndarray], pd.DataFrame, np.ndarray]


class HistBase(Track, PlotHist):
    """
    HistBase track class

    Parameters
    ----------
    style : str, optional
        Track graph type, format {'line', 'fill', 'heatmap', 'scatter'}

    fmt : str, optional
        Line styles used when the plot style is `line`, will be passed as `fmt` parameter in `matplotlib.pyplot.plot`

    line_width : int, optional
        Value used when the plot style is `line`, will be passed as `linewidth` parameter in `matplotlib.pyplot.plot`

    size : int, optional
        Value used when the plot style is `scatter`, will be passed as `s` parameter in `matplotlib.pyplot.scatter`

    color : str, optional
        Main color

    threshold_color : str, optional
        Track's color for values greater than specified threshold.

    threshold : float, optional
        Threshold used when the plot style is `line` or `scatter`, values greater than this thresh will be
        plotted with `color = threshold_color`

    cmap: str, optional
        Cmap used when the plot type is heatmap, will be passed as `cmap` paramerter in `matplotlib.pyplot.matshow`

    alpha : float, optional
        Alpha value of plot, default 1.0

    orientation : str, optional
        Track orientation, use 'inverted' for inverted track plot.

    data_range_style : {bool, 'text', 'y-axis'}, optional
        Show_data_range or not, default True.

    max_value : {float, 'auto'}, optional
        Max value of track. 'auto' for specify max value automatically, default 'auto'.

    min_value : {float, 'auto'}, optional
        Min value of track. 'auto' for specify max value automatically, default 'auto'.
    """

    STYLE_LINE = "line"
    STYLE_FILL = "fill"
    STYLE_HEATMAP = "heatmap"
    STYLE_SCATTER = "scatter"

    STYLES = (STYLE_LINE, STYLE_FILL, STYLE_HEATMAP, STYLE_SCATTER)

    DEFAULT_PROPERTIES = {
        "style": STYLE_LINE,
        "fmt": "-",
        "line_width": 2.0,
        "size": 10,
        "color": "#a6cee3",
        "threshold_color": "#ff9c9c",
        "threshold": "inf",
        "cmap": "bwr",
        "alpha": 1.0,
        "orientation": None,
        "data_range_style": "y-axis",
        "min_value": "auto",
        "max_value": "auto"
    }

    def __init__(self, **kwargs):
        properties = HistBase.DEFAULT_PROPERTIES.copy()
        properties.update({
            'type': properties['style'],
            **kwargs,
        })
        super().__init__(properties)
        self.genome_range = [None, None]

    def plot(self, ax, gr: GenomeRange, **kwargs):
        log.debug("plotting {}".format(self.properties.get('file', None)))

        self.ax = ax
        # if genome range would change?
        self.genome_range = gr

        data = self.fetch_plot_data(gr, **kwargs)
        if isinstance(data, pd.DataFrame):
            # ['pos', 'score']
            if 'pos' in data:
                data = data['pos'].to_numpy(), data['score'].to_numpy()
            # ['score']
            else:
                data = np.linspace(gr.start, gr.end, len(data)), data['score'].to_numpy()
        elif isinstance(data, np.ndarray):
            # 1d or 2d ndarray
            if len(data.shape) == 1:
                data = np.linspace(gr.start, gr.end, len(data)), data
            elif len(data.shape) == 2:
                data = np.linspace(gr.start, gr.end, data.shape[1]), data
            else:
                raise ValueError("The ndarray must be in 1d or 2d format")
        elif not isinstance(data, tuple):
            # not (indexes, values)
            raise ValueError(f"Data format not supported. should be one of {HistData}")

        indexes, values = data
        self.plot_hist(ax, gr, indexes, values)
        self.plot_label()

    def fetch_plot_data(self, gr: GenomeRange, **kwargs) -> HistData:
        """

        Returns
        -------
        hist_data: Union[Tuple[np.ndarray, np.ndarray], pd.DataFrame, np.ndarray]
            data used for plotting.
        """
        return self.fetch_data(gr, **kwargs)

    def fetch_data(self, gr: GenomeRange, **kwargs) -> HistData:
        raise NotImplementedError



"""
format convert
"""

import collections
from .filetool import opener, to_string

_fields_rg = ("bin", "name", "chrom", "strand", "txStart", "txEnd",
              "cdsStart", "cdsEnd", "exonCount", "exonStart", "exonEnds",
              "score", "name2", "cdsStartStat", "cdsEndStat", "exonFrames")

_refGeneRec = collections.namedtuple("refGeneRec", _fields_rg)


class refGeneRec(_refGeneRec):

    def to_bed12_line(self):
        chrom = self.chrom
        start = self.txStart
        end = self.txEnd
        name = self.name2
        score = self.score
        strand = self.strand

        thick_start = start
        thick_end = end

        item_rgb = "0"

        block_count = self.exonCount
        block_starts = self.offset_zero(self.exonStart, start) + ","
        block_ends = self.exonEnds
        block_sizes = self.get_exons_size(self.exonStart, block_ends) + ","

        bed12_item = (chrom, start, end, name, score, strand, thick_start, thick_end, item_rgb,
                      block_count, block_sizes, block_starts)

        bed12_line = "\t".join(bed12_item)
        return bed12_line

    def to_line(self):
        return "\t".join(self)

    @staticmethod
    def offset_zero(block, offset):
        offset = int(offset)
        block = [int(i) for i in block.split(",") if i]
        block = [str(i - offset) for i in block]
        return ",".join(block)

    @staticmethod
    def get_exons_size(exons_start, exons_end):
        starts = exons_start.split(",")
        starts = [int(i) for i in starts if i]
        ends = exons_end.split(",")
        ends = [int(i) for i in ends if i]
        sizes = [ends[i] - starts[i] for i, _ in enumerate(starts)]
        exons_size = ",".join([str(i) for i in sizes])
        return exons_size


def refgene_txt_to_bed12(txt_file, bed_file):
    with opener(txt_file) as f_in, open(bed_file, 'w') as f_out:
        for line in f_in:
            line = to_string(line)
            items = line.strip().split("\t")
            refg_rec = refGeneRec._make(items)
            out_line = refg_rec.to_bed12_line() + "\n"
            f_out.write(out_line)



import pandas as pd
from coolbox.utilities.genome import GenomeRange
from coolbox.utilities.bed import pairix_query


class FetchParix(object):
    def fetch_intervals(self, bgz_file, gr: GenomeRange, gr2: GenomeRange = None) -> pd.DataFrame:
        open_region = self.properties.get("open_region") == "yes"
        rows = list(pairix_query(bgz_file, gr, second=gr2, open_region=open_region, split=True))
        if len(rows) == 0:
            gr.change_chrom_names()
            if gr2:
                gr2.change_chrom_names()
            rows = list(pairix_query(bgz_file, gr, second=gr2, open_region=open_region, split=True))

        return pd.DataFrame(rows)



import os
import os.path as osp
import subprocess as subp
from typing import Tuple
import runpy

import fire
import nbformat as nbf

import coolbox
from coolbox.utilities import get_logger
from coolbox.utilities.genome import GenomeRange
from coolbox.api import *

log = get_logger("CoolBox CLI")


def get_element_type_by_str(elem_str):
    import coolbox.api
    try:
        elem_tp = eval("coolbox.api." + elem_str)
    except NameError:
        log.error(
            f"No element type name as {elem_tp}, all elements type see: " +
            "https://gangcaolab.github.io/CoolBox/api.html"
        )
    return elem_tp


def get_compose_code(elem_str, args, kwargs):
    compose_code = elem_str + "("
    compose_code += ", ".join(
        [repr(arg) for arg in args] +
        [f"{k} = {repr(v)}" for k, v in kwargs.items()]
    )
    compose_code += ")"
    return compose_code


FRAME_POS = ("left", "right", "top", "bottom", "center")


class CLI(object):
    """
    CoolBox Command Line Interface

    You can use this cli to create coolbox browser instance,
    visualize your data directly in shell.

    example:

    1. Draw tracks within a genome range, save figure to a pdf file:

        $ coolbox add XAxis - add BigWig test.bw - goto "chr1:5000000-6000000" - plot test.pdf

    2. Generate a notebook and run jupyter to open browser:

        $ coolbox add XAxis - add BigWig test.bw - goto "chr1:5000000-6000000" - run_jupyter

    3. Run a independent web application.

        $ coolbox add XAxis - add BigWig test.bw - goto "chr1:5000000-6000000" - run_webapp

    """

    def __init__(self, genome="hg19", genome_range: str = None, genome_range2: str = None):
        self._indent = 0
        self.current_range = [None, None]
        self.goto(genome_range, genome_range2)
        self.genome = genome
        self.frame_pos = None
        self.frames = {}

    @staticmethod
    def version():
        """print coolbox version"""
        print(coolbox.__version__)

    def set_genome(self, genome):
        """Set reference genome for browser object.

        :param genome: Reference genome (hg19, hg38, mm9, mm10),
        or the path to chromosomes size file(tab splited chromname-size).
        """
        self.genome = genome
        return self

    def goto(self, genome_range: str = None, genome_range2: str = None):
        """Goto a genome range.

        :param genome_range: Genome range string, like "chr9:4000000-6000000".
        :param genome_range2: Genome range string, like "chr9:4000000-6000000. Only required in JointView mode".
        """
        if genome_range2 is None:
            genome_range2 = genome_range
        # validate genome range
        if genome_range:
            gr = GenomeRange(genome_range)
        if genome_range2:
            gr = GenomeRange(genome_range2)
        log.info(f"Goto genome range '{genome_range}' '{genome_range2}'")
        self.current_range = [genome_range, genome_range2]

        return self

    @staticmethod
    def show_doc(elem_str):
        """Print the document of specified Element type. For example: coolbox show_doc Cool"""
        elem_tp = get_element_type_by_str(elem_str)
        print(elem_tp.__doc__)

    def joint_view(self, frame_pos: str = "top"):
        """Start a new frame positioned at the specified frame_pos in the final joint view.
        The center frame should be a single Cool, HicMat, DotHic track.

        For example:
        coolbox - \
        joint_view top - \
            add BigWig BW_PATH - \
        joint_view right - \
            add GTF BTF_PATH - \
        joint_view center - \
            add Cool - \
        goto 'chr1:100000-200000' 'chr2:200000-30000' - \
        plot /tmp/test_joint_view.svg

        Parameters
        ----------
        frame_pos: str
        Add a frame in the given position in the joint view.
        Should be one of 'top', 'left', 'center', 'bottom', 'right'.

        Returns
        -------

        """
        if frame_pos not in FRAME_POS:
            raise ValueError(f"Frame position should be one of {FRAME_POS}")
        if frame_pos not in self.frames:
            self.frames[frame_pos] = "frame = Frame()\n"
        self.frame_pos = frame_pos

        return self

    def add(self, elem_str, *args, **kwargs):
        """Add a Element(Track, Coverage, Feature), for example: coolbox add XAxis

        :param elem_str: Element type string. Like BAM, BigWig, Cool ...
        Full list of Track types
        can be found here(https://gangcaolab.github.io/CoolBox/quick_start_API.html#Track-types).
        :param args: Positional args for create elements.
        :param kwargs: Keyword args for create elements.
        """
        if ("help" in args) or ("help" in kwargs):
            self.show_doc(elem_str)
            return
        if self.frame_pos is None:
            self.joint_view("top")

        compose_code = get_compose_code(elem_str, args, kwargs)
        log.info(f"Create element, compose code: {compose_code}")
        self.frames[self.frame_pos] += "    " * self._indent + "frame += " + compose_code + "\n"
        return self

    def start_with(self, elem_str, *args, **kwargs):
        """Start a 'with' block, apply the element to all elements within the block.

        :param elem_str: Element type string. Like VLines, Color, MinValue ...
        :param args: Positional args for create elements.
        :param kwargs: Keyword args for create elements.
        """
        if ("help" in args) or ("help" in kwargs):
            self.show_doc(elem_str)
            return
        if self.frame_pos is None:
            self.joint_view("top")

        compose_code = get_compose_code(elem_str, args, kwargs)
        log.info(f"Create a with block, compose code: {compose_code}")
        self.frames[self.frame_pos] += "    " * self._indent + f"with {compose_code}:\n"
        self._indent += 1
        return self

    def end_with(self):
        """End the with block"""
        if self._indent <= 0:
            raise ValueError("Expect a 'start_with' command before end_with clause.")
        self._indent -= 1
        return self

    @staticmethod
    def _fetch_frame_src(pos: str, source: str) -> Tuple[str, str]:
        """

        Parameters
        ----------
        pos
        source

        Returns
        -------
        Tuple[str, str].
        The first string is the frame's variable name, and the second string is the function's source code
        """
        # center frame return the first track. expected a Cool, DotHiC, HicMat
        if pos != 'center':
            source += "return frame\n"
        else:
            source += "return list(frame.tracks.values())[0]\n"
        source = "\n".join("    " + line for line in source.split("\n")) + "\n"
        # generate function and call
        frame_var = f"{pos}_frame"
        source = f"def fetch_{frame_var}():\n" + source
        source += f"{frame_var} = fetch_{frame_var}()\n"

        return frame_var, source

    def source(self) -> str:
        num_frames = len(self.frames)
        if num_frames == 0:
            raise RuntimeError("No frame has been added yet.")

        if num_frames > 1 and 'center' not in self.frames:
            raise RuntimeError("JointView mode needs a center frame. "
                               "Use `joint_view center ADD xxx` to add center track/frame.")
        gr1, gr2 = self.current_range
        if gr1 is None:
            raise RuntimeError("No genome range found."
                               "Use `goto chr1:5000000-6000000` to set the genome range.")

        frame_dict = {}
        source = ""
        for pos, src in self.frames.items():
            frame_var, frame_src = self._fetch_frame_src(pos, src)
            source += frame_src
            frame_dict[pos] = frame_var

        if 'center' in self.frames:
            source += f"frame = JointView({frame_dict['center']}, " \
                      f"left={frame_dict.get('left')}, " \
                      f"right={frame_dict.get('right')}, " \
                      f"bottom={frame_dict.get('bottom')}, " \
                      f"top={frame_dict.get('top')}" \
                      f")\n"
        else:
            source += f"frame = {list(frame_dict.values())[0]}\n"

        return source

    def print_source(self):
        """Print the browser composing code."""
        print(self.source())
        return self

    def gen_notebook(self, notebook_path, notes=True, figsave=True):
        """Generate The notebook contain codes for run coolbox browser.

        :param notebook_path: The output notebook path.
        :param notes: Generate markdown notes in notebook or not.
        :param figsave: Generate codes for saving figure or not.
        """
        code = nbf.v4.new_code_cell
        markdown = nbf.v4.new_markdown_cell
        nb = nbf.v4.new_notebook()
        cells = []
        if notes:
            cells.append(
                markdown("**Run the cell bellow to start the CoolBox browser.**")
            )
        cells.append(
            code(
                f"import os; os.chdir('{os.getcwd()}')\n"
                "import coolbox.api\n" +
                "from coolbox.api import *\n" +
                self.source() +
                f"bsr = Browser(frame, reference_genome='{self.genome}')\n" +
                (f"bsr.goto('{str(self.current_range[0])}')\n" if self.current_range[0] else "") +
                "bsr.show()"
            ),
        )
        if figsave:
            if notes:
                cells.append(
                    markdown("**Run the cell bellow to save your figure.**"),
                )
            cells.append(
                code("bsr.save('test.pdf')"),
            )
        nb['cells'] = cells
        nbf.write(nb, notebook_path)
        return self

    def run_jupyter(self, jupyter_args="--ip=0.0.0.0"):
        """Create a notebook according to command line, then start a jupyter process.

        :param jupyter_args: Arguments for run jupyter.
        """
        i = 0
        tmp_notebook = lambda: f"/tmp/coolbox_tmp.{i}.ipynb"
        while osp.exists(tmp_notebook()):
            i += 1
        tmp = tmp_notebook()
        self.gen_notebook(tmp)
        subp.check_call(f"jupyter notebook {tmp} " + jupyter_args, shell=True)

    def plot(self, fig_path, genome_range=None, genome_range2=None):
        """Draw a figure within a genome range and save to file

        :param fig_path: Figure save path
        :param genome_range: Genome range string, like "chr9:4000000-6000000".
        :param genome_range2: Genome range string, like "chr9:4000000-6000000".
        """
        if genome_range is None:
            if self.current_range[0] is None:
                raise ValueError("Should specify the gr")
        else:
            self.goto(genome_range, genome_range2)
        source = "from coolbox.api import *\n" + self.source() + "\n"
        gr1, gr2 = self.current_range
        if 'center' in self.frames:
            source += f"fig = frame.plot('{gr1}', '{gr2}')\n"
            # TODO: svgutils.compose.Figure can only save to svg, convert it?
            if not fig_path.endswith('.svg'):
                fig_path = fig_path[:-4] + '.svg'
                log.warning(f"The JointView only support save to svg now. Save fig to: {fig_path}.")
            source += f"fig.save('{fig_path}')\n"
        else:
            source += f"fig = frame.plot('{gr1}')\n"
            source += f"fig.savefig('{fig_path}')\n"
        try:
            code = compile(source, "coolbox_cli_source", "exec")
            eval(code)
        except Exception as e:
            log.error(
                "Error when execute the generated source code:\n\n" +
                "------------------------\n" +
                source + "\n" +
                "------------------------\n\n"
            )
            if type(e) == NameError:
                log.error(
                    f"All elements type see: " +
                    "https://gangcaolab.github.io/CoolBox/api.html"
                )
            raise e

    def run_webapp(self, voila_args="--Voila.ip=0.0.0.0"):
        """Run a independent coolbox browser web app.
        (Create notebook and run voila)

        :param voila_args: Arguments for run jupyter.
        """
        i = 0
        tmp_notebook = lambda: f"/tmp/coolbox_tmp.{i}.ipynb"
        while osp.exists(tmp_notebook()):
            i += 1
        tmp = tmp_notebook()
        self.gen_notebook(tmp, notes=False, figsave=False)
        subp.check_call(f"voila {tmp} " + voila_args, shell=True)

    def load_module(self, mod_str):
        """
        Import custom tracks from a module/package for example:

            $ coolbox - load_module ./my_custom.py - add XAxis - add CustomTrack - goto "chr1:5000000-6000000" - run_webapp

        :param mod_str: Path to the module.
        """
        globals().update(runpy.run_path(mod_str, init_globals=globals()))
        return self


if __name__ == "__main__":
    fire.Fire(CLI)




from typing import Union

from .base import Coverage
from coolbox.utilities.genome import GenomeRange


class HLines(Coverage):
    """
    Horizontal line coverage, for show threshold.

    Parameters
    ----------
    values : float
        Y position. A list of y value or a single value.

    percent_mode : bool
        If in percent mode, y position will equal to ymin + (ymax - ymin) * val.

    color : str, optional
        Line color, default '#1e1e1e'.

    alpha : float, optional
        Line alpha value, default 0.8.

    line_style : str, optional
        Line style, default 'dashed'.

    line_width : float, optional
        Line width, default 0.5.

    name : str, optional
        The name of thr Coverage.
    """

    def __init__(self, values, **kwargs):
        if not isinstance(values, list):
            values = [values]
        properties_dict = {
            "values": values,
            "percent_mode": False,
            "color": "#1e1e1e",
            "alpha": 0.8,
            "line_style": "dashed",
            "line_width": 1,
        }
        properties_dict.update(kwargs)
        super().__init__(properties_dict)

    def plot(self, ax, gr: GenomeRange, **kwargs):
        gr = GenomeRange(gr)
        ymin, ymax = ax.get_ylim()
        if self.properties['percent_mode'] != 'no':
            hlines_list = [ymin + val * ymax for val in self.properties['values']]
        else:
            hlines_list = self.properties['values']

        ax.hlines(hlines_list, gr.start, gr.end,
                  linestyle=self.properties['line_style'],
                  linewidth=self.properties['line_width'],
                  color=self.properties['color'],
                  alpha=self.properties['alpha'])



import pandas as pd

from coolbox.utilities import GenomeRange
from coolbox.core.track.base import Track
from .plot import PlotContacts


class ArcsBase(Track, PlotContacts):
    """
    Arcs(link) track.

    Parameters
    ----------
    style : str, optional
        Style of arcs plot: 'arcs', 'hicpeaks', default 'arcs'

    score_to_width : str, optional
        Mapping function of score to width, default: '0.5 + math.sqrt(score)'

    line_width : float, optional
        Width of arc line.

    open_region : bool, optional
        If specified to True, will fetch the contacts on side in the region,
        default True

    diameter_to_height : str, optional
        Mapping function of arc diameter(interval end - start) to height.
        You can specify to 'max_diameter' let all arcs has same height.
        default 'max_height * diameter / max_diameter'.

    orientation : str, optional
        Track orientation, use 'inverted' for inverted track plot. default None

    line_style : str, optional
        Border line style, default 'solid'

    fill : bool, optional
        Fill center or not, default False.

    fill_color : str, optional
        Fill color, use 'bed_rgb' for specify color in file,
        default 'bed_rgb'.

    fill_alpha : float, optional
        Alpha value of fill region.
        default 0.2

    side : {'upper', 'lower', 'both'}
        Plot peak in which side of the matrix.
        NOTE: This parameters is useful only if the Cool track in matrix format.

    alpha : float, optional
        Alpha value of track, default 0.8.
    """

    STYLE_ARCS = 'arcs'
    STYLE_HICPEAKS = "hicpeaks"

    DEFAULT_PROPERTIES = {
        'style': STYLE_ARCS,
        # arcs
        'line_width': None,
        'score_to_width': '0.5 + math.sqrt(score)',
        'diameter_to_height': '(max_height - 0.5) * diameter / max_diameter',
        'orientation': None,
        'cmap': None,
        'vmin': None,
        'vmax': None,
        # hicpeaks
        "side": "both",
        # general
        "line_style": "solid",
        'open_region': True,
        'color': '#3297dc',
        'alpha': 0.8,
        "fill": False,
        "fill_color": "#3297dc",
        "fill_alpha": 0.2,
    }

    def __init__(self, **kwargs):
        properties = ArcsBase.DEFAULT_PROPERTIES.copy()
        properties.update(kwargs)
        super().__init__(properties)

    def fetch_plot_data(self, gr: GenomeRange, **kwargs) -> pd.DataFrame:
        """

        Returns
        -------
        intervals : pandas.core.frame.DataFrame
            Can be two types:
            1: with columns: ['pos1', 'pos2', 'score'] 'score' is optional
            2: with columns: ['start1', 'end1', 'start2', 'end2', 'score'] 'score' is optional
        """
        return self.fetch_data(gr, **kwargs)

    def plot(self, ax, gr: GenomeRange, **kwargs):
        """
        Plot arc connecting two positions on a linear scale representing interactions between bins.
        """
        self.ax = ax
        df = self.fetch_plot_data(gr, **kwargs)
        self.plot_contacts(ax, gr, kwargs.get("gr2"), df)
        self.plot_label()



import re
from numpydoc.docscrape import NumpyDocString


def paste_doc(lookup_dict):
    """
    Insert variable into obj's docstring.
    use `${val}` to insert a variable
    """

    def inner(obj):
        old_doc = obj.__doc__
        doc = ""
        old_e = 0
        for match in re.finditer("\${.*?}", old_doc):
            s, e = match.start(), match.end()
            doc += old_doc[old_e:s]
            content = old_doc[s + 2:e - 1]
            val = lookup_dict.get(content, "")
            doc += val
            old_e = e
        doc += old_doc[old_e:]
        obj.__doc__ = doc
        return obj

    return inner


class DocInheritorBase(type):
    """

    References
    ----------
    https://github.com/rsokl/custom_inherit/blob/master/src/custom_inherit/_metaclass_base.py#L17
    """

    def __new__(mcs, class_name, class_bases, class_dict, *args, **kwargs):
        from abc import abstractproperty
        # for class doc
        __doc__ = class_dict.get("__doc__", None)
        for p_cls in class_bases:
            p_cls_doc = p_cls.__doc__
            __doc__ = mcs.class_doc_inherit(p_cls_doc, __doc__)
        class_dict["__doc__"] = __doc__

        # for attr doc
        for attr, attribute in class_dict.items():
            if not mcs.inherit_attr(attr, attribute):
                continue
            child_attr = attribute
            # for static/class methods, the __doc__ is unchangeable
            if isinstance(child_attr, (staticmethod, classmethod)):
                child_attr = child_attr.__func__

            prnt_attr_doc = None
            # Even for middle cls with no attr, the hasattr will find the attr recursively.
            for p_cls in (cls for cls in class_bases if hasattr(cls, attr)):
                prnt_attr_doc = getattr(p_cls, attr).__doc__
                if prnt_attr_doc is not None:
                    break

            if prnt_attr_doc is None:
                continue

            doc = mcs.attr_doc_inherit(prnt_attr_doc, child_attr.__doc__)
            try:
                child_attr.__doc__ = doc
            # property.__doc__ is read-only in Python 2 (TypeError), 3.3 - 3.4 (AttributeError)
            except (TypeError, AttributeError) as err:
                if type(child_attr) in (property, abstractproperty):
                    new_prop = property(
                        fget=child_attr.fget,
                        fset=child_attr.fset,
                        fdel=child_attr.fdel,
                        doc=doc,
                    )
                    if isinstance(child_attr, abstractproperty):
                        new_prop = abstractproperty(new_prop)
                    class_dict[attr] = new_prop
                else:
                    raise type(err)(err)

        return type.__new__(mcs, class_name, class_bases, class_dict)

    @classmethod
    def class_doc_inherit(mcs, p_doc, c_doc) -> str:
        return c_doc

    @classmethod
    def attr_doc_inherit(mcs, p_doc, c_doc) -> str:
        return c_doc

    @classmethod
    def inherit_attr(mcs, attr_name, attr):
        from types import FunctionType, MethodType
        # inherit docstring for method, static-method, class-method, abstract-method, decorated-method, and property
        return isinstance(attr, (FunctionType, MethodType, classmethod, staticmethod, property))


class NumpyDocInheritor(DocInheritorBase):

    @classmethod
    def build(mcs, p_doc: str, c_doc: str) -> str:
        if not p_doc or not c_doc:
            return p_doc or c_doc

        p_doc = NumpyDocString(p_doc)
        c_doc = NumpyDocString(c_doc)

        # reuse parents' doc except for `Extended Summary`
        for section, content in c_doc.items():
            if section != "Extended Summary" and (not content and p_doc[section]):
                c_doc[section] = p_doc[section]

        # merge parameters
        c_params = [param for param in c_doc['Parameters']]
        c_param_names = set(param.name for param in c_params)
        p_params = [param for param in p_doc['Parameters'] if param.name not in c_param_names]
        c_params += p_params

        c_doc['Parameters'] = c_params

        return str(c_doc)

    @classmethod
    def class_doc_inherit(mcs, p_doc, c_doc) -> str:
        return mcs.build(p_doc, c_doc)

    @classmethod
    def attr_doc_inherit(mcs, p_doc, c_doc) -> str:
        return mcs.build(p_doc, c_doc)



from .bedpe import BEDPE
from .base import ArcsBase


class HiCPeaks(BEDPE):
    """
    Hi-C Peaks(Loops) from .bedpe file. Used to show the peaks on the Hi-C interaction map.
    """
    DEFAULT_PROPERTIES = {
        "style": ArcsBase.STYLE_HICPEAKS,
        'color': "#2255ff",
        "open_region": True,
        "alpha": 0.6,
        "line_width": 5,
    }

    def __init__(self, file, **kwargs):
        properties = HiCPeaks.DEFAULT_PROPERTIES.copy()
        properties.update(kwargs)
        super().__init__(file, **properties)



import pandas as pd

from coolbox.utilities import (
    get_logger
)
from coolbox.utilities.genome import GenomeRange

from coolbox.core.track.base import Track
from .plot import PlotBed

log = get_logger(__name__)


class BedBase(Track, PlotBed):
    """
    BED Base track.

    Parameters
    ----------
    style : {'gene', 'tad'}

    gene_style: {'flybase', 'normal'}

    display : {'stacked', 'interlaced', 'collapsed'}, optional
        Display mode. (Default: 'stacked')

    color : str, optional
        Track color, 'bed_rgb' for auto specify color according to bed record.
        (Default: 'bed_rgb')

    border_color : str, optional
        Border_color of gene. (Default: 'black')

    fontsize : int, optional
        Font size. (Default: BED.DEFAULT_FONTSIZE)

    labels : {True, False, 'auto'}, optional
        Draw bed name or not. 'auto' for automate decision according to density.
        (Default: 'auto')

    interval_height : int, optional
        The height of the interval. (Default: 100)

    num_rows : int, optional
        Set the max interval rows. (Default: unlimited interval rows)

    max_value : float, optional
        Max score. (Default: inf)

    min_value : float, optional
        Min score. (Default: -inf)

    border_style: str, optional
        Border style of tad. (Default: 'solid')

    border_width: int, optional
        Border width of tad. (Default: '2.0')

    show_score : bool
        Show bed score or not.
        default False.

    score_font_size : {'auto', int}
        Score text font size.
        default 'auto'

    score_font_color : str
        Score text color.
        default '#000000'

    score_height_ratio : float
        (text tag height) / (TAD height). used for adjust the position of Score text.
        default 0.5

    border_only : bool
        Only show border, default False

    """

    STYLE_GENE = "gene"
    STYLE_TAD = "tad"

    COLOR = "#1f78b4"

    DEFAULT_PROPERTIES = {
        'style': STYLE_GENE,
        # gene
        'gene_style': 'flybase',
        'display': 'stacked',
        'color': "bed_rgb",
        'border_color': "#1f78b4",
        'fontsize': 12,
        'interval_height': 100,
        'num_rows': None,
        'labels': 'off',
        'min_score': '-inf',
        'max_score': 'inf',
        'bed_type': None,
        # tad
        'border_style': "--",
        'border_width': 2.0,
        "show_score": False,
        "score_font_size": 'auto',
        "score_font_color": "#000000",
        "score_height_ratio": 0.4,
        "border_only": False,
    }

    def __init__(self, **kwargs):
        properties = BedBase.DEFAULT_PROPERTIES.copy()
        properties.update(kwargs)
        super().__init__(properties)
        self.init_for_plot()

    def fetch_data(self, gr: GenomeRange, **kwargs) -> pd.DataFrame:
        """

        Returns
        -------
        intervals : pandas.core.frame.DataFrame
            BED interval table. The table should be in format like:

            bed_fields = ['chromosome', 'start', 'end',
                          'name', 'score', 'strand',
                          'thick_start', 'thick_end',
                          'rgb', 'block_count',
                          'block_sizes', 'block_starts']
            The table can be in bed6/bed9/bed12 format and the trailing columns can be omited.

        """
        raise NotImplementedError

    def plot(self, ax, gr: GenomeRange, **kwargs):
        self.ax = ax
        ov_intervals: pd.DataFrame = self.fetch_plot_data(gr, **kwargs)

        style = self.properties['style']
        if style == self.STYLE_TAD:
            self.plot_tads(ax, gr, ov_intervals)
        elif style == self.STYLE_GENE:
            self.plot_genes(ax, gr, ov_intervals)
        else:
            raise ValueError("style not supportted, should be one of 'gene' 'tad' ")
        self.plot_label()



from copy import copy

from coolbox.utilities import (
    op_err_msg, get_feature_stack,
    get_coverage_stack, format_properties,
    get_logger
)
from coolbox.utilities.genome import GenomeRange
from coolbox.utilities.doctool import NumpyDocInheritor


log = get_logger(__name__)


class Track(object, metaclass=NumpyDocInheritor):
    """
    Track base class.

    Parameters
    ----------
    properties_dict : dict
        The properties(features) of this track. For example 'height', 'color'...

    name : str, optional
        The name of Track.
        (Default: "{self.__class__.__name__}.{self.__class__._counts}")

    title : str, optional
        Title of ax

    height : int, optional
        Height of ax

    color : str, optional;
        Color of ax

    Attributes
    ----------
    properties : dict
        The properties(features) of this track. For example 'height', 'color'...

    name : str
        The name of Track.

    coverages : list of `coolbox.api.coverage.Coverage`
        Coverages on this Track.
    """

    DEFAULT_PROPERTIES = {
        'height': 2,
        "color": "#808080",
        "name": "",
        "title": "",
    }

    counts = 0
    tracks = {}

    def __new__(cls, *args, **kwargs):
        cls.counts += 1
        # An environment to record all exist tracks
        return super().__new__(cls)

    def __init__(self, properties_dict={}, **kwargs):
        # TODO replace properties_dict with **kwargs ? Then the initialization in all classes are unified
        properties = Track.DEFAULT_PROPERTIES.copy()
        properties.update(properties_dict)
        properties.update(kwargs)
        self.properties = format_properties(properties)

        name = self.properties.get('name')
        if name:
            assert isinstance(name, str), "Track name must be a `str`."
        else:
            name = self.__class__.__name__ + ".{}".format(self.__class__.counts)
        Track.tracks[name] = self
        self.properties['name'] = name
        self.name = name

        # disable call mixin class
        # super().__init__()
        self.coverages = []

        # load features from global feature stack
        features_stack = get_feature_stack()
        for features in features_stack:
            self.properties.update(features.properties)

        # load coverages from global coverages stack
        coverage_stack = get_coverage_stack()
        for coverage in coverage_stack:
            self.coverages.append(coverage)

        self.ax = None

    def __del__(self):
        name = self.properties['name']
        if name in Track.tracks:
            del Track.tracks[name]

    def fetch_data(self, gr: GenomeRange, **kwargs):
        """
        Fetch the raw data for the given GenomeRange.

        Parameters
        ----------
        gr: GenomeRange
        kwargs: dict
            Other optional parameters including `gr2`, `resolution` for some tracks.

        Returns
        -------

        """
        raise NotImplementedError

    def fetch_plot_data(self, gr: GenomeRange, **kwargs):
        """
        Fetch the plot data for the given GenomeRange.
        The returned data would be used in `plot` function, the default implementation directly return `fetch_data`.

        Parameters
        ----------
        gr: GenomeRange
            GenomeRange is an object with properties of [chrom, start, end].

        kwargs: dict
            Other optional parameters including `gr2`, `resolution` for some tracks.

        Returns
        -------

        """
        return self.fetch_data(gr, **kwargs)

    def plot(self, ax, gr: GenomeRange, **kwargs):
        """
        Plot figure based on data returned by `fetch_plot_data`.

        Parameters
        ----------
        gr: GenomeRange
        kwargs: dict
            Other optional parameters including `gr2`, `resolution` for some tracks.

        Returns
        -------

        """
        raise NotImplementedError

    @property
    def name(self):
        return self.properties['name']

    @name.setter
    def name(self, value):
        self.properties['name'] = value

    def __add__(self, other):
        from ..frame.base import FrameBase
        from ..frame.frame import Frame
        from ..coverage.base import Coverage
        from ..coverage.base import CoverageStack
        from ..feature import Feature

        if isinstance(other, Track):
            result = Frame()
            result.add_track(self)
            result.add_track(other)
            return result
        elif isinstance(other, FrameBase):
            result = copy(other)
            result.add_track(self, pos='head')
            return result
        elif isinstance(other, Feature):
            result = copy(self)
            return other.__add__(result)
        elif isinstance(other, Coverage):
            result = copy(self)
            result.append_coverage(other)
            return result
        elif isinstance(other, CoverageStack):
            result = copy(self)
            result.pile_coverages(other.coverages, pos='top')
            return result
        else:
            raise TypeError(op_err_msg(self, other))

    def append_coverage(self, coverage, pos='top'):
        """
        Append coverage to this track.

        Parameters
        ----------
        coverage : `coolbox.api.coverage.Coverage`
            Coverage object to be piled.

        pos : {'top', 'bottom'}, optional
            Add coverages to top or bottom. (Default: 'top')
        """
        if pos == 'top':
            self.coverages.append(coverage)
        else:
            self.coverages.insert(0, coverage)

    def pile_coverages(self, coverages, pos='top'):
        """
        Pile a stack of coverages with self's coverages

        Parameters
        ----------
        coverages : list of `coolbox.api.coverage.Coverage` or `coolbox.api.coverage.CoverageStack`
            Coverage objects to be piled.

        pos : {'top', 'bottom'}, optional
            Add coverages to top or bottom. (Default: 'top')
        """
        from ..coverage.base import CoverageStack

        if isinstance(coverages, CoverageStack):
            coverages = coverages.coverages
        elif isinstance(coverages, list):
            pass
        else:
            raise TypeError("coverages must a list of Coverage or CoverageStack")

        if not hasattr(self, 'coverages'):
            self.coverages = coverages
        else:
            if pos == 'top':
                self.coverages.extend(coverages)
            else:
                self.coverages = coverages + self.coverages

    def has_prop(self, p):
        return (p in self.properties) and (self.properties[p] is not None)

    def plot_label(self):
        if hasattr(self, 'label_ax') and self.label_ax is not None:
            self.label_ax.text(0.15, 0.5, self.properties['title'],
                               horizontalalignment='left', size='large',
                               verticalalignment='center')

    def plot_coverages(self, ax, gr: GenomeRange, gr2: GenomeRange):
        """
        Plot all coverages on given axes.

        Parameters
        ----------
        gr : GenomeRange
            First genome range.

        gr2 : GenomeRange
            Second genome range.
        """
        if not hasattr(self, 'coverages'):
            return
        for cov_idx, cov in enumerate(self.coverages):
            cov.track = self
            if hasattr(cov, 'track_instance'):
                cov.track_instance.track = self
            try:
                cov.plot(ax, copy(gr), gr2=copy(gr2))
            except Exception as e:
                log.error("Error occured when plot track's coverage:\n"
                          "\ttrack name: {}\n\ttrack type:{}\n\tcoverage name: {}\n\tcov type: {}\n"
                          "\tError: {} {}".format(
                            self.name, type(self), cov.name, type(cov),
                            type(e), str(e)))
                log.exception(e)



import abc
import re
from typing import Tuple

import numpy as np
from scipy.linalg import toeplitz
from scipy.ndimage import gaussian_filter
from scipy.signal import convolve2d

from coolbox.utilities.logtools import get_logger

log = get_logger(__name__)


class ProcessHicMat(object):
    SMALL_VALUE = 1e-12

    def process_matrix(self, arr: np.ndarray) -> np.ndarray:
        properties = self.properties
        transform = properties['transform']
        normalize = properties['normalize']
        gaussian_sigma = properties['gaussian_sigma']
        process_func = properties['process_func']
        # process the matrix
        if transform != 'no':
            arr = self.transform_matrix(arr, transform)
        if normalize != 'no':
            arr = self.normalize_matrix(arr, normalize)
        if gaussian_sigma != 'no':
            arr = self.gaussian_matrix(arr, gaussian_sigma)
        if process_func != 'no':
            # user-defined process function
            func = process_func
            try:
                if callable(func):
                    arr = func(arr)
                elif isinstance(func, str):
                    func = eval(func)
                    arr = func(arr)
                else:
                    raise ValueError("process_func")
            except Exception as e:
                log.error(str(e))
                raise ValueError(
                    "process_func should a one argument function "
                    "receive a matrix return a processed matrix.")
        return arr

    def fill_zero_nan(self, arr: np.ndarray) -> np.ndarray:
        # fill zero and nan with small value
        small = self.SMALL_VALUE
        zero_indices = arr == 0
        nan_indices = np.isnan(arr)
        self.zero_indices = zero_indices
        self.nan_indices = nan_indices
        arr[zero_indices] = small
        arr[nan_indices] = small
        return arr

    @staticmethod
    def transform_matrix(arr: np.ndarray, method='log') -> np.ndarray:
        if method == 'log10':
            arr = np.log10(arr)
        elif method == 'log2':
            arr = np.log2(arr)
        elif method == 'log':
            arr = np.log(arr)
        return arr

    @staticmethod
    def gaussian_matrix(arr: np.ndarray, sigma=1.0) -> np.ndarray:
        arr = gaussian_filter(arr, sigma)
        return arr

    @classmethod
    def normalize_matrix(cls, mat: np.ndarray, cis=True, method='zscore') -> np.ndarray:
        if cis:
            res = cls.normalize_cis(mat, method)
        else:
            res = cls.normalize_trans(mat, method)
        return res

    @classmethod
    def normalize_cis(cls, mat: np.ndarray, method: str) -> np.ndarray:
        res = mat
        if method == 'total':
            total = np.sum(mat)
            if total != 0:
                res = mat / total
        elif method == 'expect':
            means = cls.diagonal_mean(mat)
            expect = toeplitz(means)
            res = mat / expect
        elif method == 'zscore':
            means, stds = cls.diagonal_mean_std(mat)
            mat_mean = toeplitz(means)
            mat_std = toeplitz(stds)
            res = (mat - mat_mean) / mat_std
        elif re.match("hiccups:.:.", method):
            p, w = method.strip("hiccups:").split(":")
            p, w = int(w), int(p)
            kernel = cls.__donut_kernel(p, w)

            def apply_donut(m):
                m_ext = np.zeros((m.shape[0] + 2 * (w - 1), m.shape[0] + 2 * (w - 1)))
                idx_center = slice(w, w + m.shape[0]), slice(w, w + m.shape[1])
                m_ext[idx_center] = m
                m_f = convolve2d(m_ext, kernel, mode='same')
                m_f = m_f[idx_center]
                return m_f

            means = cls.diagonal_mean(mat)
            exp_decay = toeplitz(means)
            m_donut = apply_donut(mat)
            exp_donut = apply_donut(exp_decay)
            exp = (m_donut / exp_donut) * exp_decay
            res = mat / exp
        else:
            log.warning(f"Cis-matrix does not support the {method} normalize.")
        return res

    @classmethod
    def normalize_trans(cls, mat: np.ndarray, method: str) -> np.ndarray:
        res = mat
        if method == 'total':
            total = np.sum(mat)
            if total != 0:
                res = mat / total
        elif method == 'zscore':
            mean = np.mean(mat)
            std = np.std(mat)
            res = (mat - mean) / std
        else:
            log.warning(f"Trans-matrix doest not support the {method} normalize.")
        return res

    @staticmethod
    def diagonal_mean(mat: np.ndarray) -> np.ndarray:
        return np.array([np.diagonal(mat, i).mean() for i in range(mat.shape[0])])

    @staticmethod
    def diagonal_mean_std(mat: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        means = []
        stds = []
        for i in range(mat.shape[0]):
            diagonal = np.diagonal(mat, i)
            means.append(diagonal.mean())
            stds.append(diagonal.std())
        stds = np.array(stds)
        try:
            stds[stds == 0] = stds[stds > 0].min()
        except Exception:
            log.error("Error when executing 'stds[stds > 0].min()'")
            pass
        return np.array(means), stds


    @staticmethod
    def __donut_kernel(p: int, w: int) -> np.ndarray:
        k1 = np.ones((2 * w + 1, 2 * w + 1))
        k2 = np.zeros((2 * w + 1, 2 * w + 1))
        k2[w - p:w + p + 1, w - p:w + p + 1] = 1
        k3 = np.zeros((2 * w + 1, 2 * w + 1))
        k3[:w - p, w] = 1
        k3[w, :w - p] = 1
        k3[w, w + 2 * p - 1:2 * w + 1] = 1
        k3[w + 2 * p - 1:2 * w + 1, w] = 1
        k = k1 - k2 - k3
        return k



import typing as t
from os.path import abspath, dirname, join
from collections import deque

from .bed import *
from .figtools import *
from .filetool import *
from .fmtconvert import *
from .genome import *
from .logtools import *


def op_err_msg(a, b, op='+'):
    """
    Generate the error message of error operand type.
    """
    msg = "unsupported operand type(s) for {}: '{}' and '{}'".format(op, str(type(a)), str(type(b)))
    return msg


#
# read genome length file #

here = dirname(abspath(__file__))

HG19 = GenomeLength(join(here, "../genome/hg19.txt"), genome_name="hg19")
HG38 = GenomeLength(join(here, "../genome/hg38.txt"), genome_name="hg38")
MM9 = GenomeLength(join(here, "../genome/mm9.txt"), genome_name="mm9")
MM10 = GenomeLength(join(here, "../genome/mm10.txt"), genome_name="mm10")

BUILT_IN_GENOMES = {
    'hg19': HG19,
    'hg38': HG38,
    'mm9': MM9,
    'mm10': MM10,
}

FEATURES_STACK_NAME = "__COOLBOX_FEATURE_STACK__"
COVERAGE_STACK_NAME = "__COOLBOX_COVERAGE_STACK__"
IGNORE_TYPE_CHANGE = ['title','name']

def get_feature_stack():
    global_scope = globals()
    global_scope.setdefault(FEATURES_STACK_NAME, deque())
    return global_scope[FEATURES_STACK_NAME]


def get_coverage_stack():
    global_scope = globals()
    global_scope.setdefault(COVERAGE_STACK_NAME, deque())
    return global_scope[COVERAGE_STACK_NAME]


def format_properties(properties):
    """
    Convert bool value to 'yes' or 'no', for compatible with pyGenomeTracks
    Convert string to float if possible
    """
    for key, value in properties.items():
        if isinstance(value, bool):
            properties[key] = 'yes' if value else 'no'
        elif isinstance(value, str):
            if key in IGNORE_TYPE_CHANGE:
                continue
            try:
                float_val = float(value)
                properties[key] = float_val
            except Exception:
                pass
    return properties


def correspond_track(track_or_name: t.Union["Track", str]):
    """
    Search track by it's name.
    For mark corresponding track in a element definition, allow
    pass track name to find related object in CLI mode.
    """
    from coolbox.core.track.base import Track
    if isinstance(track_or_name, str):
        track = Track.tracks[track_or_name]
    else:
        track = track_or_name
    return track


if __name__ == "__main__":
    import doctest

    doctest.testmod()



import numpy as np

from coolbox.utilities import to_gr
from .base import HicMatBase


class DotHiC(HicMatBase):
    """
    HicMat track from .hic file.

    Parameters
    ----------
    file: str
        The file path of .hic file.

    balance : {bool, 'KR', 'VC', 'VC_SQRT'}, optional
        Matrix balance method,
        default True('KR' balance)

    """
    DEFAULT_PROPERTIES = {
        'cmap': "JuiceBoxLike2",
        'balance': True,
        "norm": "no",
        "transform": "log"
    }

    def __init__(self, file, **kwargs):
        properties = DotHiC.DEFAULT_PROPERTIES.copy()
        properties.update({
            'file': file,
            **kwargs
        })
        super().__init__(**properties)

    def fetch_data(self, genome_range, genome_range2=None, **kwargs) -> np.ndarray:
        from coolbox.utilities.hic.wrap import StrawWrap

        path = self.properties['file']
        wrap = StrawWrap(path, normalization=self.balance, binsize=kwargs.get('resolution', 'auto'))

        arr = wrap.fetch(genome_range, genome_range2)

        self.fetched_binsize = wrap.fetched_binsize  # expose fetched binsize

        return self.fill_zero_nan(arr)

    def fetch_pixels(self, gr, gr2=None, balance=None, **kwargs):
        """

        Parameters
        ----------
        gr : {str, GenomeRange}
            Intervals within input chromosome range.

        gr2 : {str, GenomeRange}
            Intervals within input chromsome range2.

        balance : {bool, 'KR', 'VC', 'VC_SQRT'}, optional
            matrix balance method,
            default `self.balance`.

        resolution : {'auto', int}
            resolution of the data. for example 5000.
            'auto' for calculate resolution automatically.
            default 'auto'

        Returns
        -------
        pixels : pandas.core.frame.DataFrame
            Hi-C pixels table.
            The pixel table contains the non-zero upper triangle entries of the contact map.
        """
        from coolbox.utilities.hic.wrap import StrawWrap

        gr = to_gr(gr)
        if gr2 is not None:
            gr2 = to_gr(gr2)

        path = self.properties['file']
        balance = kwargs.get('balance', self.is_balance)
        wrap = StrawWrap(path, normalization=balance, binsize=kwargs.get('resolution', 'auto'))

        pixels = wrap.fetch_pixels(gr, gr2)
        return pixels

    def infer_binsize(self, genome_range1, genome_range2=None, **kwargs) -> int:
        from coolbox.utilities.hic.wrap import StrawWrap

        path = self.properties['file']
        wrap = StrawWrap(path, normalization=self.balance, binsize=kwargs.get('resolution', 'auto'))
        gr1 = to_gr(genome_range1)
        return wrap.infer_binsize(gr1)
